"""StateStore — persistence layer for agent/pipeline state."""
from __future__ import annotations

from dataclasses import asdict, is_dataclass
from datetime import UTC, datetime
import hashlib
import json
from pathlib import Path
import sqlite3
from typing import Any

from logos_adk.config import StateConfig
from logos_adk.errors import ConfigError, StateBackendError, StateError
from logos_adk.persistence import (
    get_postgresql_schema_version,
    get_sqlite_schema_version,
    set_postgresql_schema_version,
    set_sqlite_schema_version,
    sqlite_table_columns,
)
from logos_adk.postgres_utils import pooled_connect, resolve_database_url, safe_identifier
from logos_adk.state import AgentState
from logos_adk.types import Message, ToolCall, ToolResult, normalize_content_parts


STATE_SCHEMA_KIND = "state_store"
STATE_SCHEMA_VERSION = 1


def _json_default(value: Any) -> Any:
    if isinstance(value, datetime):
        return value.isoformat()
    if is_dataclass(value):
        return asdict(value)
    raise TypeError(f"Object of type {type(value).__name__} is not JSON serializable")


def _serialize_state_data(state: AgentState) -> dict[str, Any]:
    return {
        "messages": [
            {
                "role": message.role,
                "content": message.content,
                "content_parts": message.content_parts,
                "sender": message.sender,
                "tool_calls": [
                    {
                        "id": tool_call.id,
                        "name": tool_call.name,
                        "arguments": tool_call.arguments,
                    }
                    for tool_call in message.tool_calls
                ],
                "tool_result": (
                    {
                        "call_id": message.tool_result.call_id,
                        "content": message.tool_result.content,
                        "content_parts": message.tool_result.content_parts,
                        "error": message.tool_result.error,
                        "duration_ms": message.tool_result.duration_ms,
                        "status": message.tool_result.status,
                    }
                    if message.tool_result is not None
                    else None
                ),
            }
            for message in state.messages
        ],
        "session_id": state.session_id,
        "metadata": state.metadata,
        "version": state.version,
        "updated_at": state.updated_at.isoformat() if state.updated_at else None,
    }


def _deserialize_state_data(payload: dict[str, Any]) -> AgentState:
    messages = []
    for raw_message in payload.get("messages", []):
        tool_calls = [
            ToolCall(
                id=item["id"],
                name=item["name"],
                arguments=item.get("arguments", {}),
            )
            for item in raw_message.get("tool_calls", [])
        ]
        raw_tool_result = raw_message.get("tool_result")
        tool_result = None
        if raw_tool_result is not None:
            tool_result = ToolResult(
                call_id=raw_tool_result["call_id"],
                content=raw_tool_result.get("content"),
                content_parts=normalize_content_parts(raw_tool_result.get("content_parts")),
                error=raw_tool_result.get("error"),
                duration_ms=raw_tool_result.get("duration_ms", 0),
                status=raw_tool_result.get("status", "completed"),
            )
        messages.append(
            Message(
                role=raw_message["role"],
                content=raw_message.get("content", ""),
                content_parts=normalize_content_parts(raw_message.get("content_parts")),
                sender=raw_message.get("sender"),
                tool_calls=tool_calls,
                tool_result=tool_result,
            )
        )

    updated_at = payload.get("updated_at")
    return AgentState(
        messages=messages,
        session_id=payload.get("session_id"),
        metadata=payload.get("metadata", {}),
        version=int(payload.get("version", 0)),
        updated_at=datetime.fromisoformat(updated_at) if updated_at else None,
    )


def _clone_state(state: AgentState) -> AgentState:
    return _deserialize_state_data(
        json.loads(json.dumps(_serialize_state_data(state), default=_json_default))
    )


class StateStore:
    """Abstract base for state persistence."""

    def save(self, namespace: str, state: AgentState) -> AgentState:
        raise NotImplementedError

    def load(self, namespace: str, session_id: str | None) -> AgentState | None:
        raise NotImplementedError

    def history(self, namespace: str, session_id: str | None, *, limit: int = 50) -> list[dict[str, Any]]:
        raise NotImplementedError

    def export_state(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> dict[str, Any] | None:
        raise NotImplementedError

    def import_state(
        self,
        namespace: str,
        payload: dict[str, Any],
        *,
        session_id: str | None = None,
    ) -> AgentState:
        raise NotImplementedError

    def rollback(self, namespace: str, session_id: str | None, *, version: int) -> AgentState:
        raise NotImplementedError

    def describe(self) -> dict[str, Any]:
        raise NotImplementedError


class InMemoryStateStore(StateStore):
    """In-memory state store for testing and development."""

    def __init__(self) -> None:
        self._latest: dict[tuple[str, str], AgentState] = {}
        self._history: dict[tuple[str, str], list[AgentState]] = {}

    def _key(self, namespace: str, session_id: str | None) -> tuple[str, str]:
        return (namespace, session_id or "")

    def save(self, namespace: str, state: AgentState) -> AgentState:
        key = self._key(namespace, state.session_id)
        stored_state = _clone_state(state)
        stored_state.version = len(self._history.get(key, [])) + 1
        stored_state.updated_at = datetime.now(UTC)
        self._latest[key] = stored_state
        self._history.setdefault(key, []).append(_clone_state(stored_state))
        return _clone_state(stored_state)

    def load(self, namespace: str, session_id: str | None) -> AgentState | None:
        state = self._latest.get(self._key(namespace, session_id))
        return _clone_state(state) if state is not None else None

    def history(self, namespace: str, session_id: str | None, *, limit: int = 50) -> list[dict[str, Any]]:
        versions = list(reversed(self._history.get(self._key(namespace, session_id), [])))[:limit]
        return [
            {
                "version": item.version,
                "saved_at": item.updated_at.isoformat() if item.updated_at else "",
                "session_id": item.session_id,
            }
            for item in versions
        ]

    def export_state(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> dict[str, Any] | None:
        versions = self._history.get(self._key(namespace, session_id), [])
        if not versions:
            return None
        if version is None:
            state = versions[-1]
        else:
            state = next((item for item in versions if item.version == version), None)
            if state is None:
                return None
        return {
            "agent_name": namespace,
            "session_id": state.session_id,
            "version": state.version,
            "saved_at": state.updated_at.isoformat() if state.updated_at else "",
            "state": _serialize_state_data(state),
        }

    def import_state(
        self,
        namespace: str,
        payload: dict[str, Any],
        *,
        session_id: str | None = None,
    ) -> AgentState:
        state_payload = payload.get("state", payload)
        state = _deserialize_state_data(state_payload)
        state.session_id = session_id if session_id is not None else state.session_id
        return self.save(namespace, state)

    def rollback(self, namespace: str, session_id: str | None, *, version: int) -> AgentState:
        versions = self._history.get(self._key(namespace, session_id), [])
        restored = next((item for item in versions if item.version == version), None)
        if restored is None:
            raise StateError(f"No saved state version {version} for session {session_id}")
        rolled_back = _clone_state(restored)
        rolled_back.session_id = session_id
        return self.save(namespace, rolled_back)

    def describe(self) -> dict[str, Any]:
        return {"backend": "in_memory", "shards": 1}


class SQLiteStateStore(StateStore):
    """Persistent state store backed by SQLite."""

    def __init__(self, path: str, table_name: str = "agent_state_versions") -> None:
        self.path = path
        self.table_name = safe_identifier(table_name)
        self._schema_ready = False

    def _connect(self) -> sqlite3.Connection:
        try:
            Path(self.path).parent.mkdir(parents=True, exist_ok=True)
            connection = sqlite3.connect(self.path)
            connection.row_factory = sqlite3.Row
            return connection
        except Exception as exc:
            raise StateBackendError(f"Failed to open SQLite state backend: {exc}") from exc

    def _ensure_schema(self, connection: sqlite3.Connection) -> None:
        if self._schema_ready:
            return
        connection.execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self.table_name} (
                namespace TEXT NOT NULL,
                session_id TEXT NOT NULL,
                version INTEGER NOT NULL,
                saved_at TEXT NOT NULL,
                state_json TEXT NOT NULL,
                PRIMARY KEY (namespace, session_id, version)
            )
            """
        )
        columns = self._get_columns(connection)
        if {"agent_name", "session_key", "session_id", "version", "saved_at", "state_json"}.issubset(columns):
            self._migrate_legacy_schema(connection)
            columns = self._get_columns(connection)
        if {"namespace", "session_id", "version", "saved_at", "state_json"}.issubset(columns):
            connection.execute(
                f"""
                CREATE INDEX IF NOT EXISTS idx_{self.table_name}_latest
                ON {self.table_name} (namespace, session_id, version DESC)
                """
            )
        recorded_version = get_sqlite_schema_version(connection, STATE_SCHEMA_KIND)
        if recorded_version != STATE_SCHEMA_VERSION:
            set_sqlite_schema_version(connection, STATE_SCHEMA_KIND, STATE_SCHEMA_VERSION)
        connection.commit()
        self._schema_ready = True

    def _session_key(self, session_id: str | None) -> str:
        return session_id or ""

    def _get_columns(self, connection: sqlite3.Connection) -> set[str]:
        return sqlite_table_columns(connection, self.table_name)

    def _uses_legacy_schema(self, connection: sqlite3.Connection) -> bool:
        columns = self._get_columns(connection)
        return "namespace" not in columns and "agent_name" in columns and "session_key" in columns

    def _migrate_legacy_schema(self, connection: sqlite3.Connection) -> None:
        legacy_table = f"{self.table_name}__legacy_v0"
        connection.execute(f"ALTER TABLE {self.table_name} RENAME TO {legacy_table}")
        connection.execute(
            f"""
            CREATE TABLE {self.table_name} (
                namespace TEXT NOT NULL,
                session_id TEXT NOT NULL,
                version INTEGER NOT NULL,
                saved_at TEXT NOT NULL,
                state_json TEXT NOT NULL,
                PRIMARY KEY (namespace, session_id, version)
            )
            """
        )
        connection.execute(
            f"""
            INSERT INTO {self.table_name} (namespace, session_id, version, saved_at, state_json)
            SELECT agent_name, session_key, version, saved_at, state_json
            FROM {legacy_table}
            """
        )
        connection.execute(f"DROP TABLE {legacy_table}")

    def save(self, namespace: str, state: AgentState) -> AgentState:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                session_id = self._session_key(state.session_id)
                current_version = connection.execute(
                    f"""
                    SELECT COALESCE(MAX(version), 0)
                    FROM {self.table_name}
                    WHERE namespace = ? AND session_id = ?
                    """,
                    (namespace, session_id),
                ).fetchone()[0]
                stored_state = _clone_state(state)
                stored_state.version = int(current_version) + 1
                stored_state.updated_at = datetime.now(UTC)
                payload = json.dumps(_serialize_state_data(stored_state), default=_json_default)
                connection.execute(
                    f"""
                    INSERT INTO {self.table_name} (namespace, session_id, version, saved_at, state_json)
                    VALUES (?, ?, ?, ?, ?)
                    """,
                    (
                        namespace,
                        session_id,
                        stored_state.version,
                        stored_state.updated_at.isoformat(),
                        payload,
                    ),
                )
                connection.commit()
                return _clone_state(stored_state)
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to persist state: {exc}") from exc

    def _fetch_state_row(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> sqlite3.Row | None:
        with self._connect() as connection:
            self._ensure_schema(connection)
            session_key = self._session_key(session_id)
            if version is None:
                return connection.execute(
                    f"""
                    SELECT version, saved_at, state_json
                    FROM {self.table_name}
                    WHERE namespace = ? AND session_id = ?
                    ORDER BY version DESC
                    LIMIT 1
                    """,
                    (namespace, session_key),
                ).fetchone()
            return connection.execute(
                f"""
                SELECT version, saved_at, state_json
                FROM {self.table_name}
                WHERE namespace = ? AND session_id = ? AND version = ?
                LIMIT 1
                """,
                (namespace, session_key, version),
            ).fetchone()

    def load(self, namespace: str, session_id: str | None) -> AgentState | None:
        try:
            row = self._fetch_state_row(namespace, session_id)
            if row is None:
                return None
            state = _deserialize_state_data(json.loads(row["state_json"]))
            state.version = int(row["version"])
            state.updated_at = datetime.fromisoformat(row["saved_at"])
            return state
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to load state: {exc}") from exc

    def history(self, namespace: str, session_id: str | None, *, limit: int = 50) -> list[dict[str, Any]]:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                rows = connection.execute(
                    f"""
                    SELECT version, saved_at
                    FROM {self.table_name}
                    WHERE namespace = ? AND session_id = ?
                    ORDER BY version DESC
                    LIMIT ?
                    """,
                    (namespace, self._session_key(session_id), limit),
                ).fetchall()
            return [
                {
                    "version": int(row["version"]),
                    "saved_at": row["saved_at"],
                    "session_id": session_id,
                }
                for row in rows
            ]
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to list state history: {exc}") from exc

    def export_state(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> dict[str, Any] | None:
        try:
            row = self._fetch_state_row(namespace, session_id, version=version)
            if row is None:
                return None
            return {
                "agent_name": namespace,
                "session_id": session_id,
                "version": int(row["version"]),
                "saved_at": row["saved_at"],
                "state": json.loads(row["state_json"]),
            }
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to export state: {exc}") from exc

    def import_state(
        self,
        namespace: str,
        payload: dict[str, Any],
        *,
        session_id: str | None = None,
    ) -> AgentState:
        try:
            state_payload = payload.get("state", payload)
            state = _deserialize_state_data(state_payload)
            state.session_id = session_id if session_id is not None else state.session_id
            return self.save(namespace, state)
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to import state: {exc}") from exc

    def rollback(self, namespace: str, session_id: str | None, *, version: int) -> AgentState:
        exported = self.export_state(namespace, session_id, version=version)
        if exported is None:
            raise StateError(f"No saved state version {version} for session {session_id}")
        return self.import_state(namespace, exported["state"], session_id=session_id)

    def describe(self) -> dict[str, Any]:
        backend = "sqlite"
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                if self._uses_legacy_schema(connection):
                    backend = "sqlite_legacy"
        except Exception:
            pass
        return {
            "backend": backend,
            "path": self.path,
            "table_name": self.table_name,
            "schema_version": STATE_SCHEMA_VERSION,
            "shards": 1,
        }


class PostgreSQLStateStore(StateStore):
    """Persistent state store backed by PostgreSQL."""

    def __init__(self, dsn: str, table_name: str = "agent_state_versions") -> None:
        self.dsn = dsn
        self.table_name = safe_identifier(table_name)
        self._schema_ready = False

    def _connect(self) -> Any:
        try:
            return pooled_connect(self.dsn)
        except Exception as exc:
            raise StateBackendError(f"Failed to open PostgreSQL state backend: {exc}") from exc

    def _ensure_schema(self, connection: Any) -> None:
        if self._schema_ready:
            return
        with connection.cursor() as cursor:
            cursor.execute(
                f"""
                CREATE TABLE IF NOT EXISTS {self.table_name} (
                    namespace TEXT NOT NULL,
                    session_id TEXT NOT NULL,
                    version INTEGER NOT NULL,
                    saved_at TIMESTAMPTZ NOT NULL,
                    state_json JSONB NOT NULL,
                    PRIMARY KEY (namespace, session_id, version)
                )
                """
            )
            cursor.execute(
                f"""
                CREATE INDEX IF NOT EXISTS idx_{self.table_name}_latest
                ON {self.table_name} (namespace, session_id, version DESC)
                """
            )
        recorded_version = get_postgresql_schema_version(connection, STATE_SCHEMA_KIND)
        if recorded_version != STATE_SCHEMA_VERSION:
            set_postgresql_schema_version(connection, STATE_SCHEMA_KIND, STATE_SCHEMA_VERSION)
        connection.commit()
        self._schema_ready = True

    def _session_key(self, session_id: str | None) -> str:
        return session_id or ""

    def save(self, namespace: str, state: AgentState) -> AgentState:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                session_id = self._session_key(state.session_id)
                with connection.cursor() as cursor:
                    cursor.execute(
                        f"""
                        SELECT COALESCE(MAX(version), 0)
                        FROM {self.table_name}
                        WHERE namespace = %s AND session_id = %s
                        """,
                        (namespace, session_id),
                    )
                    current_version = int(cursor.fetchone()[0])
                    stored_state = _clone_state(state)
                    stored_state.version = current_version + 1
                    stored_state.updated_at = datetime.now(UTC)
                    cursor.execute(
                        f"""
                        INSERT INTO {self.table_name} (namespace, session_id, version, saved_at, state_json)
                        VALUES (%s, %s, %s, %s, %s::jsonb)
                        """,
                        (
                            namespace,
                            session_id,
                            stored_state.version,
                            stored_state.updated_at,
                            json.dumps(_serialize_state_data(stored_state), default=_json_default),
                        ),
                    )
                connection.commit()
                return _clone_state(stored_state)
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to persist state: {exc}") from exc

    def load(self, namespace: str, session_id: str | None) -> AgentState | None:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                with connection.cursor() as cursor:
                    cursor.execute(
                        f"""
                        SELECT version, saved_at, state_json
                        FROM {self.table_name}
                        WHERE namespace = %s AND session_id = %s
                        ORDER BY version DESC
                        LIMIT 1
                        """,
                        (namespace, self._session_key(session_id)),
                    )
                    row = cursor.fetchone()
            if row is None:
                return None
            state = _deserialize_state_data(row[2] if isinstance(row[2], dict) else json.loads(row[2]))
            state.version = int(row[0])
            state.updated_at = row[1] if isinstance(row[1], datetime) else datetime.fromisoformat(row[1])
            return state
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to load state: {exc}") from exc

    def history(self, namespace: str, session_id: str | None, *, limit: int = 50) -> list[dict[str, Any]]:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                with connection.cursor() as cursor:
                    cursor.execute(
                        f"""
                        SELECT version, saved_at
                        FROM {self.table_name}
                        WHERE namespace = %s AND session_id = %s
                        ORDER BY version DESC
                        LIMIT %s
                        """,
                        (namespace, self._session_key(session_id), limit),
                    )
                    rows = cursor.fetchall()
            return [
                {
                    "version": int(row[0]),
                    "saved_at": row[1].isoformat() if isinstance(row[1], datetime) else str(row[1]),
                    "session_id": session_id,
                }
                for row in rows
            ]
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to list state history: {exc}") from exc

    def export_state(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> dict[str, Any] | None:
        try:
            with self._connect() as connection:
                self._ensure_schema(connection)
                with connection.cursor() as cursor:
                    if version is None:
                        cursor.execute(
                            f"""
                            SELECT version, saved_at, state_json
                            FROM {self.table_name}
                            WHERE namespace = %s AND session_id = %s
                            ORDER BY version DESC
                            LIMIT 1
                            """,
                            (namespace, self._session_key(session_id)),
                        )
                    else:
                        cursor.execute(
                            f"""
                            SELECT version, saved_at, state_json
                            FROM {self.table_name}
                            WHERE namespace = %s AND session_id = %s AND version = %s
                            LIMIT 1
                            """,
                            (namespace, self._session_key(session_id), version),
                        )
                    row = cursor.fetchone()
            if row is None:
                return None
            state_json = row[2] if isinstance(row[2], dict) else json.loads(row[2])
            return {
                "agent_name": namespace,
                "session_id": session_id,
                "version": int(row[0]),
                "saved_at": row[1].isoformat() if isinstance(row[1], datetime) else str(row[1]),
                "state": state_json,
            }
        except StateBackendError:
            raise
        except Exception as exc:
            raise StateBackendError(f"Failed to export state: {exc}") from exc

    def import_state(
        self,
        namespace: str,
        payload: dict[str, Any],
        *,
        session_id: str | None = None,
    ) -> AgentState:
        state_payload = payload.get("state", payload)
        state = _deserialize_state_data(state_payload)
        state.session_id = session_id if session_id is not None else state.session_id
        return self.save(namespace, state)

    def rollback(self, namespace: str, session_id: str | None, *, version: int) -> AgentState:
        exported = self.export_state(namespace, session_id, version=version)
        if exported is None:
            raise StateError(f"No saved state version {version} for session {session_id}")
        return self.import_state(namespace, exported["state"], session_id=session_id)

    def describe(self) -> dict[str, Any]:
        return {
            "backend": "postgresql",
            "dsn": self.dsn,
            "table_name": self.table_name,
            "schema_version": STATE_SCHEMA_VERSION,
            "shards": 1,
        }


class ShardedStateStore(StateStore):
    """Routes state operations across multiple backend shards."""

    def __init__(self, shards: list[StateStore]) -> None:
        if not shards:
            raise ValueError("ShardedStateStore requires at least one shard")
        self._shards = shards

    def _shard_index(self, key: str) -> int:
        digest = hashlib.sha256(key.encode("utf-8")).hexdigest()
        return int(digest, 16) % len(self._shards)

    def _select_shard(self, session_id: str | None) -> StateStore:
        return self._shards[self._shard_index(session_id or "__default__")]

    def save(self, namespace: str, state: AgentState) -> AgentState:
        return self._select_shard(state.session_id).save(namespace, state)

    def load(self, namespace: str, session_id: str | None) -> AgentState | None:
        return self._select_shard(session_id).load(namespace, session_id)

    def history(self, namespace: str, session_id: str | None, *, limit: int = 50) -> list[dict[str, Any]]:
        return self._select_shard(session_id).history(namespace, session_id, limit=limit)

    def export_state(
        self,
        namespace: str,
        session_id: str | None,
        *,
        version: int | None = None,
    ) -> dict[str, Any] | None:
        return self._select_shard(session_id).export_state(namespace, session_id, version=version)

    def import_state(
        self,
        namespace: str,
        payload: dict[str, Any],
        *,
        session_id: str | None = None,
    ) -> AgentState:
        resolved_session_id = session_id
        if resolved_session_id is None:
            state_payload = payload.get("state", payload)
            resolved_session_id = state_payload.get("session_id")
        return self._select_shard(resolved_session_id).import_state(
            namespace,
            payload,
            session_id=session_id,
        )

    def rollback(self, namespace: str, session_id: str | None, *, version: int) -> AgentState:
        return self._select_shard(session_id).rollback(namespace, session_id, version=version)

    def describe(self) -> dict[str, Any]:
        return {
            "backend": "sharded",
            "shards": len(self._shards),
            "stores": [shard.describe() for shard in self._shards],
        }


def build_state_store(
    config: StateConfig | None,
    *,
    default_persistent: bool = False,
) -> StateStore:
    """Build a configured state store backend."""
    state_config = config or StateConfig()
    database_url = resolve_database_url(state_config.dsn)
    backend_name = state_config.backend

    if backend_name is None:
        if default_persistent:
            backend_name = "postgresql" if database_url else "sqlite"
        else:
            backend_name = "in_memory"

    if backend_name in {"in_memory", "memory", "inmemory"}:
        return InMemoryStateStore()
    if backend_name == "sqlite":
        if not state_config.path:
            raise ConfigError("SQLite state backend requires a non-empty path")
        return SQLiteStateStore(state_config.path, table_name=state_config.table_name)
    if backend_name in {"postgres", "postgresql"}:
        if not database_url:
            raise ConfigError(
                "PostgreSQL state backend requires a DSN or DATABASE_URL/LOGOS_ADK_DATABASE_URL"
            )
        return PostgreSQLStateStore(database_url, table_name=state_config.table_name)
    if backend_name == "sharded":
        if not state_config.shards:
            raise ConfigError("Sharded state backend requires a non-empty 'shards' list")
        return ShardedStateStore([build_state_store(shard, default_persistent=True) for shard in state_config.shards])

    raise ConfigError(f"Unsupported state backend: {backend_name}")


__all__ = [
    "StateStore",
    "InMemoryStateStore",
    "SQLiteStateStore",
    "PostgreSQLStateStore",
    "ShardedStateStore",
    "build_state_store",
]



"""Source document management."""
from __future__ import annotations

import json
import sqlite3
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Any

from .converters import row_to_source_document
from .models import SourceDocumentSnapshot

if TYPE_CHECKING:
    from .engine import RetrievalLearningEngine


class SourceDocumentsMixin:
    """Mixin for source document functionality."""

    @staticmethod
    def register_source_document(
        engine: "RetrievalLearningEngine",
        *,
        workspace_id: str | None,
        source_key: str,
        document_id: str,
        content: str,
        metadata: dict[str, Any],
        source_ref: str | None = None,
    ) -> SourceDocumentSnapshot:
        """Register a source document snapshot."""
        snapshot = SourceDocumentSnapshot(
            workspace_id=workspace_id,
            source_key=source_key,
            document_id=document_id,
            content=content,
            metadata=dict(metadata),
            source_ref=source_ref,
        )
        with engine._connect() as connection:
            engine._ensure_schema(connection)
            if engine.dsn:
                connection.execute(
                    f"""
                    INSERT INTO {engine._table('source_documents')}
                    (workspace_id, source_key, document_id, content, metadata_json, source_ref, updated_at)
                    VALUES (?, ?, ?, ?, ?, ?, ?)
                    ON CONFLICT (workspace_id, source_key, document_id)
                    DO UPDATE SET
                        content = EXCLUDED.content,
                        metadata_json = EXCLUDED.metadata_json,
                        source_ref = EXCLUDED.source_ref,
                        updated_at = EXCLUDED.updated_at
                    """,
                    (
                        engine._workspace_value(workspace_id),
                        source_key,
                        document_id,
                        content,
                        json.dumps(snapshot.metadata, sort_keys=True),
                        source_ref,
                        snapshot.updated_at,
                    ),
                )
            else:
                connection.execute(
                    f"""
                    INSERT OR REPLACE INTO {engine._table('source_documents')}
                    (workspace_id, source_key, document_id, content, metadata_json, source_ref, updated_at)
                    VALUES (?, ?, ?, ?, ?, ?, ?)
                    """,
                    (
                        engine._workspace_value(workspace_id),
                        source_key,
                        document_id,
                        content,
                        json.dumps(snapshot.metadata, sort_keys=True),
                        source_ref,
                        snapshot.updated_at,
                    ),
                )
                SourceDocumentsMixin._prune_source_documents(
                    connection,
                    engine=engine,
                    workspace_key=engine._workspace_value(workspace_id),
                    source_key=source_key,
                )
            connection.commit()
        return snapshot

    @staticmethod
    def list_source_documents(
        engine: "RetrievalLearningEngine",
        *,
        workspace_id: str | None = None,
        source_key: str | None = None,
    ) -> list[SourceDocumentSnapshot]:
        """List source document snapshots."""
        clauses = ["workspace_id = ?"]
        params: list[Any] = [engine._workspace_value(workspace_id)]
        if source_key is not None:
            clauses.append("source_key = ?")
            params.append(source_key)
        with engine._connect() as connection:
            engine._ensure_schema(connection)
            rows = connection.execute(
                f"""
                SELECT * FROM {engine._table('source_documents')}
                WHERE {' AND '.join(clauses)}
                ORDER BY updated_at DESC
                """,
                params,
            ).fetchall()
        return [row_to_source_document(row, workspace_id) for row in rows]

    @staticmethod
    def prune_source_documents(
        engine: "RetrievalLearningEngine",
        *,
        workspace_id: str | None = None,
        source_key: str | None = None,
        ttl_days: int | None = None,
        max_documents_per_source: int | None = None,
    ) -> int:
        """Prune old source documents."""
        deleted = 0
        with engine._connect() as connection:
            engine._ensure_schema(connection)
            deleted += SourceDocumentsMixin._prune_source_documents(
                connection,
                engine=engine,
                workspace_key=engine._workspace_value(workspace_id) if workspace_id is not None else None,
                source_key=source_key,
                ttl_days=ttl_days,
                max_documents_per_source=max_documents_per_source,
            )
            connection.commit()
        return deleted

    @staticmethod
    def _prune_source_documents(
        connection: sqlite3.Connection,
        *,
        engine: "RetrievalLearningEngine",
        workspace_key: str | None,
        source_key: str | None,
        ttl_days: int | None = None,
        max_documents_per_source: int | None = None,
    ) -> int:
        """Internal prune source documents."""
        deleted = 0
        effective_ttl_days = ttl_days if ttl_days is not None else engine.source_document_ttl_days
        effective_max = max_documents_per_source if max_documents_per_source is not None else engine.max_source_documents_per_source
        
        clauses: list[str] = []
        params: list[Any] = []
        if workspace_key is not None:
            clauses.append("workspace_id = ?")
            params.append(workspace_key)
        if source_key is not None:
            clauses.append("source_key = ?")
            params.append(source_key)
        
        where_sql = f"WHERE {' AND '.join(clauses)}" if clauses else ""
        cutoff = (datetime.now(UTC) - timedelta(days=effective_ttl_days)).isoformat()
        
        # Delete stale documents
        stale_rows = connection.execute(
            f"""
            SELECT workspace_id, source_key, document_id
            FROM {engine._table('source_documents')}
            {where_sql} {"AND" if where_sql else "WHERE"} updated_at < ?
            """,
            [*params, cutoff],
        ).fetchall()
        for row in stale_rows:
            connection.execute(
                f"""
                DELETE FROM {engine._table('source_documents')}
                WHERE workspace_id = ? AND source_key = ? AND document_id = ?
                """,
                (row["workspace_id"], row["source_key"], row["document_id"]),
            )
            deleted += 1

        # Enforce max documents per source
        grouped_rows = connection.execute(
            f"""
            SELECT workspace_id, source_key, document_id
            FROM {engine._table('source_documents')}
            {where_sql}
            ORDER BY workspace_id ASC, source_key ASC, updated_at DESC
            """,
            params,
        ).fetchall()
        seen: dict[tuple[str, str], int] = {}
        for row in grouped_rows:
            key = (row["workspace_id"], row["source_key"])
            seen[key] = seen.get(key, 0) + 1
            if seen[key] <= effective_max:
                continue
            connection.execute(
                f"""
                DELETE FROM {engine._table('source_documents')}
                WHERE workspace_id = ? AND source_key = ? AND document_id = ?
                """,
                (row["workspace_id"], row["source_key"], row["document_id"]),
            )
            deleted += 1
        return deleted



"""Tests for Agent.from_config YAML loading."""
import os
from pathlib import Path
import pytest
from logos_adk.agent import Agent
from logos_adk.errors import ConfigError, ToolNotFoundError
from logos_adk.memory import LongTermMemory, SessionMemory, SharedMemory
from logos_adk.memory.backends import InMemoryBackend, SQLiteBackend
from logos_adk.observe import ConsoleExporter, InMemoryExporter, OpenTelemetryExporter
from logos_adk.guardrails import ContentModerationGuard, HallucinationGuard, ToxicityGuard
from logos_adk.tools import tool


FIXTURE_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


def _register_yaml_tools():
    """Register test tools in the registry for YAML loading tests."""

    @tool
    async def yaml_search(query: str) -> str:
        """Search for YAML test."""
        return query

    @tool
    async def yaml_ticket(title: str) -> dict:
        """Create ticket for YAML test."""
        return {"title": title}


def test_from_config_basic():
    _register_yaml_tools()
    path = os.path.join(FIXTURE_DIR, "test_agent.yaml")
    agent = Agent.from_config(path)
    assert agent.name == "test_bot"
    assert agent._config.model == "claude-sonnet"
    assert agent._config.system_prompt == "You are a test bot"


def test_from_config_loads_provider_proxy(tmp_path):
    config_path = tmp_path / "proxy_agent.yaml"
    config_path.write_text(
        "\n".join(
            [
                "name: proxy_bot",
                "provider:",
                "  type: openai-compatible",
                "  model: local-model",
                "  base_url: http://localhost:1234/v1",
                "  proxy: http://proxy.internal:8080",
            ]
        )
    )

    agent = Agent.from_config(str(config_path))

    assert agent._config.provider is not None
    assert agent._config.provider.proxy == "http://proxy.internal:8080"


def test_from_config_loads_tools():
    _register_yaml_tools()
    path = os.path.join(FIXTURE_DIR, "test_agent.yaml")
    agent = Agent.from_config(path)
    tool_names = [t.name for t in agent._config.tools]
    assert "yaml_search" in tool_names
    assert "yaml_ticket" in tool_names


def test_from_config_unknown_tool():
    path = os.path.join(FIXTURE_DIR, "bad_tools_agent.yaml")
    with pytest.raises(ToolNotFoundError, match="nonexistent_tool"):
        Agent.from_config(path)


def test_from_config_memory_session_shorthand():
    path = os.path.join(FIXTURE_DIR, "memory_session_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.memories) == 1
    assert isinstance(agent._config.memories[0], SessionMemory)


def test_from_config_memory_structured():
    path = os.path.join(FIXTURE_DIR, "memory_structured_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.memories) == 2
    assert isinstance(agent._config.memories[0], SessionMemory)
    assert isinstance(agent._config.memories[1], LongTermMemory)
    assert agent._config.memories[0].strategy == "summarize"
    assert agent._config.memories[0].max_messages == 6
    assert agent._config.memories[1].retrieval == "hybrid"
    assert agent._config.memories[1].limit == 3


def test_from_config_memory_sqlite_backend(tmp_path):
    config_path = tmp_path / "sqlite_agent.yaml"
    db_path = tmp_path / "memory.db"
    config_path.write_text(
        "\n".join(
            [
                "name: memory_bot",
                "model: claude-sonnet",
                "memory:",
                "  long_term:",
                "    backend: sqlite",
                f"    path: {db_path}",
                "    retrieval: keyword",
            ]
        )
    )

    agent = Agent.from_config(str(config_path))

    assert len(agent._config.memories) == 1
    assert isinstance(agent._config.memories[0], LongTermMemory)
    assert isinstance(agent._config.memories[0].backend, SQLiteBackend)
    assert Path(agent._config.memories[0].backend.path) == db_path


def test_from_config_memory_shared_backend():
    path = os.path.join(FIXTURE_DIR, "memory_shared_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.memories) == 1
    assert isinstance(agent._config.memories[0], SharedMemory)
    assert isinstance(agent._config.memories[0].backend, InMemoryBackend)
    assert agent._config.memories[0].limit == 4


def test_from_config_unknown_memory_backend():
    path = os.path.join(FIXTURE_DIR, "bad_memory_agent.yaml")

    with pytest.raises(ConfigError, match="Unsupported memory backend"):
        Agent.from_config(path)


def test_from_config_sqlite_requires_path(tmp_path):
    config_path = tmp_path / "bad_sqlite_agent.yaml"
    config_path.write_text(
        "\n".join(
            [
                "name: memory_bot",
                "model: claude-sonnet",
                "memory:",
                "  long_term:",
                "    backend: sqlite",
            ]
        )
    )

    agent = Agent.from_config(str(config_path))
    assert isinstance(agent._config.memories[0].backend, SQLiteBackend)
    assert agent._config.memories[0].backend.path.endswith("long_term_memory.db")


def test_from_config_memory_requires_backend_for_long_term():
    path = os.path.join(FIXTURE_DIR, "bad_memory_missing_backend_agent.yaml")
    agent = Agent.from_config(path)
    assert isinstance(agent._config.memories[0].backend, SQLiteBackend)


def test_from_config_unknown_memory_section():
    path = os.path.join(FIXTURE_DIR, "bad_memory_unknown_section_agent.yaml")

    with pytest.raises(ConfigError, match="Unknown memory config sections"):
        Agent.from_config(path)


def test_from_config_rejects_unknown_session_strategy():
    path = os.path.join(FIXTURE_DIR, "bad_memory_strategy_agent.yaml")

    with pytest.raises(ConfigError, match="Unsupported session memory strategy"):
        Agent.from_config(path)


def test_from_config_token_limit_requires_max_tokens():
    path = os.path.join(FIXTURE_DIR, "bad_memory_token_limit_agent.yaml")

    with pytest.raises(ConfigError, match="requires max_tokens"):
        Agent.from_config(path)


def test_from_config_rejects_bad_session_param_type():
    path = os.path.join(FIXTURE_DIR, "bad_memory_max_messages_type_agent.yaml")

    with pytest.raises(ConfigError, match="max_messages must be an integer"):
        Agent.from_config(path)


def test_from_config_rejects_bad_long_term_retrieval():
    path = os.path.join(FIXTURE_DIR, "bad_memory_retrieval_agent.yaml")

    with pytest.raises(ConfigError, match="Unsupported long-term memory retrieval"):
        Agent.from_config(path)


def test_from_config_observe_console_shorthand():
    path = os.path.join(FIXTURE_DIR, "observe_console_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], ConsoleExporter)


def test_from_config_observe_console_structured():
    path = os.path.join(FIXTURE_DIR, "observe_console_structured_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], ConsoleExporter)


def test_from_config_observe_memory_shorthand():
    path = os.path.join(FIXTURE_DIR, "observe_memory_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], InMemoryExporter)


def test_from_config_observe_memory_structured():
    path = os.path.join(FIXTURE_DIR, "observe_memory_structured_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], InMemoryExporter)


def test_from_config_observe_otel_structured():
    path = os.path.join(FIXTURE_DIR, "observe_otel_agent.yaml")
    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], OpenTelemetryExporter)
    assert agent._config.observers[0].endpoint == "http://collector.test/v1/traces"
    assert agent._config.observers[0].timeout == 7.5
    assert agent._config.observers[0].batch_size == 25
    assert agent._config.observers[0].max_retries == 4
    assert agent._config.observers[0].backoff_seconds == 0.75
    assert agent._config.observers[0].bearer_token == "secret-token"


def test_from_config_loads_guardrails_and_scaling(tmp_path):
    config_path = tmp_path / "guarded_agent.yaml"
    config_path.write_text(
        "\n".join(
            [
                "name: guarded_bot",
                "model: claude-sonnet",
                "guardrails:",
                "  input:",
                "    - type: moderation",
                "      blocked_categories: [violence]",
                "    - type: toxicity",
                "      threshold: 1",
                "  output:",
                "    - type: hallucination",
                "      required_facts: [PostgreSQL]",
                "scaling:",
                "  distributed: true",
                "  session_affinity: true",
            ]
        )
    )

    agent = Agent.from_config(str(config_path))

    assert any(isinstance(guard, ContentModerationGuard) for guard in agent._config.input_guardrails)
    assert any(isinstance(guard, ToxicityGuard) for guard in agent._config.input_guardrails)
    assert any(isinstance(guard, HallucinationGuard) for guard in agent._config.output_guardrails)
    assert agent._config.scaling.session_affinity is True


def test_from_config_observe_otel_env_resolution(monkeypatch):
    monkeypatch.setenv("OTEL_BEARER_TOKEN", "env-secret-token")
    monkeypatch.setenv("OTEL_TENANT_ID", "tenant-from-env")
    monkeypatch.setenv("OTEL_API_KEY", "api-key-from-env")
    path = os.path.join(FIXTURE_DIR, "observe_otel_env_agent.yaml")

    agent = Agent.from_config(path)

    assert len(agent._config.observers) == 1
    assert isinstance(agent._config.observers[0], OpenTelemetryExporter)
    assert agent._config.observers[0].bearer_token == "env-secret-token"
    assert agent._config.observers[0].headers == {
        "x-tenant-id": "tenant-from-env",
        "x-api-key": "api-key-from-env",
    }


def test_from_config_rejects_unsupported_observe_exporter():
    path = os.path.join(FIXTURE_DIR, "bad_observe_unknown_exporter_agent.yaml")

    with pytest.raises(ConfigError, match="Unsupported observe exporter"):
        Agent.from_config(path)


def test_from_config_rejects_observe_without_exporter():
    path = os.path.join(FIXTURE_DIR, "bad_observe_missing_exporter_agent.yaml")

    with pytest.raises(ConfigError, match="requires an 'exporter' field"):
        Agent.from_config(path)


def test_from_config_rejects_otel_without_endpoint():
    path = os.path.join(FIXTURE_DIR, "bad_observe_agent.yaml")

    with pytest.raises(ConfigError, match="requires a non-empty endpoint"):
        Agent.from_config(path)


def test_from_config_rejects_invalid_otel_batch_size():
    path = os.path.join(FIXTURE_DIR, "bad_observe_otel_batch_size_agent.yaml")

    with pytest.raises(ConfigError, match="batch_size must be a positive integer"):
        Agent.from_config(path)


def test_from_config_rejects_invalid_otel_headers():
    path = os.path.join(FIXTURE_DIR, "bad_observe_otel_headers_agent.yaml")

    with pytest.raises(ConfigError, match="headers must be a string-to-string mapping"):
        Agent.from_config(path)


def test_from_config_rejects_invalid_otel_compression():
    path = os.path.join(FIXTURE_DIR, "bad_observe_otel_compression_agent.yaml")

    with pytest.raises(ConfigError, match="compression must be 'none' or 'gzip'"):
        Agent.from_config(path)


def test_from_config_rejects_missing_otel_env(monkeypatch):
    monkeypatch.delenv("OTEL_MISSING_TOKEN", raising=False)
    path = os.path.join(FIXTURE_DIR, "bad_observe_otel_missing_env_agent.yaml")

    with pytest.raises(ConfigError, match="references environment variable 'OTEL_MISSING_TOKEN'"):
        Agent.from_config(path)



"""Hot-reload support for Logos ADK agents."""
from __future__ import annotations

import os
import threading
import time
from typing import Any, Callable

from logos_adk.errors import ConfigError


class ConfigWatcher:
    """Watches YAML config files and reloads agents when they change."""

    def __init__(
        self,
        directory: str,
        *,
        interval: float = 2.0,
        on_reload: Callable[[Any, str], None] | None = None,
        on_error: Callable[[Any, str, Exception], None] | None = None,
    ) -> None:
        self._directory = directory
        self._interval = interval
        self._on_reload = on_reload
        self._on_error = on_error
        self._watched: dict[str, tuple[float, Any]] = {}
        self._watched_lock = threading.Lock()
        self._running = False
        self._thread: threading.Thread | None = None
        self._loop: Any | None = None

    def register(self, agent: Any) -> None:
        """Register an agent for config watching.

        Raises:
            ConfigError: If the agent has no config path.
        """
        path = getattr(agent, "_config_path", None)
        if path is None:
            raise ConfigError(
                "Cannot watch agent: agent has no config path. "
                "Use Agent.from_config() to create the agent."
            )
        mtime = os.stat(path).st_mtime
        with self._watched_lock:
            self._watched[path] = (mtime, agent)

    def start(self, loop: Any | None = None) -> None:
        """Start watching for config changes in a background thread.

        Args:
            loop: Event loop for scheduling reloads via call_soon_threadsafe.
                  If None, reload() is called directly.
        """
        if self._running:
            return
        self._loop = loop
        self._running = True
        self._thread = threading.Thread(target=self._poll_loop, daemon=True)
        self._thread.start()

    def stop(self) -> None:
        """Stop the watcher."""
        self._running = False
        if self._thread is not None:
            self._thread.join(timeout=self._interval * 2)
            self._thread = None

    def _poll_loop(self) -> None:
        """Background polling loop."""
        while self._running:
            time.sleep(self._interval)
            if not self._running:
                break
            with self._watched_lock:
                watched_snapshot = list(self._watched.items())
            for path, (old_mtime, agent) in watched_snapshot:
                try:
                    current_mtime = os.stat(path).st_mtime
                except OSError:
                    continue

                if current_mtime != old_mtime:
                    try:
                        if self._loop is not None:
                            self._loop.call_soon_threadsafe(agent.reload)
                        else:
                            agent.reload()
                        with self._watched_lock:
                            self._watched[path] = (current_mtime, agent)
                        if self._on_reload is not None:
                            self._on_reload(agent, path)
                    except Exception as exc:
                        with self._watched_lock:
                            self._watched[path] = (current_mtime, agent)
                        if self._on_error is not None:
                            self._on_error(agent, path, exc)



"""MessagesMixin — message building, memory, system prompts."""
from __future__ import annotations
from typing import Any

from logos_adk.agent._constants import _UNSET
from logos_adk.memory import MemoryEntry
from logos_adk.types import Message
from logos_adk.token_counter import count_text as _count_text


class MessagesMixin:
    async def _remember(
        self,
        messages: list[Message],
        state: Any,
        session_id: str | None,
    ) -> list[MemoryEntry]:
        """Collect memory entries relevant to the current turn."""
        entries: list[MemoryEntry] = []
        for memory in self._config.memories:
            entries.extend(await memory.remember(messages, state, session_id=session_id))
        return entries

    def _memory_messages(self, entries: list[MemoryEntry]) -> list[Message]:
        """Render retrieved memories into system messages, respecting token budget."""
        if not entries:
            return []

        budget = self._config.context_window.memory_max_tokens

        if budget <= 0:
            # No budget — include all entries
            return [
                Message(
                    role="system",
                    content=(
                        "Relevant memory:\n" + "\n".join(f"- {entry.content}" for entry in entries)
                    ),
                )
            ]

        # Include entries until we'd exceed the budget
        included: list[str] = []
        used_tokens = _count_text("Relevant memory:\n", model="")
        for entry in entries:
            line = f"- {entry.content}"
            line_tokens = _count_text(line, model="")
            if used_tokens + line_tokens > budget and included:
                break
            included.append(line)
            used_tokens += line_tokens

        if not included:
            return []

        return [
            Message(
                role="system",
                content="Relevant memory:\n" + "\n".join(included),
            )
        ]

    async def _commit_memories(
        self,
        messages: list[Message],
        response: Any,
        state: Any,
        session_id: str | None,
    ) -> None:
        """Persist memory after a completed agent turn."""
        for memory in self._config.memories:
            await memory.commit(messages, response, state, session_id=session_id)

    def _normalize_prompt_input(
        self,
        prompt: str | list,
    ) -> tuple[str, list]:
        if isinstance(prompt, str):
            return prompt, []
        from logos_adk.types import normalize_content_parts, content_text
        parts = normalize_content_parts(prompt)
        return content_text("", parts), parts

    def _resolved_system_prompt(self, system_prompt_override: Any = _UNSET) -> str | None:
        # If override is provided, skip cache
        if system_prompt_override is not _UNSET:
            return self._compute_system_prompt(system_prompt_override)
        # Check run-scoped cache
        if self._run_cache_depth > 0 and self._cached_system_prompt is not _UNSET:
            return self._cached_system_prompt  # type: ignore[return-value]
        result = self._compute_system_prompt(_UNSET)
        if self._run_cache_depth > 0:
            self._cached_system_prompt = result
        return result

    def _compute_system_prompt(self, system_prompt_override: Any) -> str | None:
        base_prompt = self._config.system_prompt if system_prompt_override is _UNSET else system_prompt_override
        role_prompt = self._role_system_prompt()

        # --- Self-Learning Injection ---
        lessons = self._get_learned_lessons()
        learning_prompt = None
        if lessons:
            learning_prompt = "### LEARNED FROM PREVIOUS FEEDBACK:\n" + "\n".join(
                f"- {lesson}" for lesson in lessons
            )
        # -------------------------------
        budget_prompt = self._budget_system_prompt()

        parts = [
            part.strip()
            for part in [role_prompt, base_prompt, learning_prompt, budget_prompt]
            if isinstance(part, str) and part.strip()
        ]
        if not parts:
            return None
        return "\n\n".join(parts)

    def _get_learned_lessons(self) -> list[str]:
        """Retrieve learned lessons from agent metadata."""
        lessons = self._config.metadata.get("learned_lessons", [])
        if not isinstance(lessons, list):
            return []
        return lessons

    def learn_from_feedback(self, feedback: str) -> None:
        """Save a new lesson learned from user feedback."""
        if not feedback or len(feedback) < 3:
            return
        lessons = self._config.metadata.setdefault("learned_lessons", [])
        if feedback not in lessons:
            lessons.append(feedback)
            # Persist if needed
            if self._config_path:
                try:
                    self.save_config(self._config_path)
                except Exception:
                    pass

    def _role_system_prompt(self) -> str | None:
        profile = self.crew_profile
        if not profile:
            return None
        lines = ["Crew Profile:"]
        if "role" in profile:
            lines.append(f"Role: {profile['role']}")
        if "goal" in profile:
            lines.append(f"Goal: {profile['goal']}")
        if "backstory" in profile:
            lines.append(f"Backstory: {profile['backstory']}")
        return "\n".join(lines)



"""Direct tests for file-based agent config loading."""
from __future__ import annotations

import textwrap

import pytest

from logos_adk.agent import Agent
from logos_adk.agent_registry import agent_class_registry, agent_factory_registry
from logos_adk.config import AgentConfig
from logos_adk.config_loader import build_agent_config, load_agent_config, load_agent_from_config
from logos_adk.errors import ConfigError, ToolNotFoundError
from logos_adk.memory import SessionMemory
from logos_adk.observe import ConsoleExporter
from logos_adk.tools import tool
from logos_adk.tools.registry import tool_registry


@pytest.fixture(autouse=True)
def _clear_tool_registry():
    tool_registry.clear()
    yield
    tool_registry.clear()


@pytest.fixture(autouse=True)
def _clear_agent_class_registry():
    agent_class_registry.clear()
    agent_factory_registry.clear()
    yield
    agent_class_registry.clear()
    agent_factory_registry.clear()


class RegistryLoadedAgent(Agent):
    pass


class ImportPathAgent(Agent):
    pass


class ModuleLoadedAgent(Agent):
    pass


class FactoryLoadedAgent(Agent):
    pass


class NoArgFactoryAgent(Agent):
    pass


def _register_loader_tools():
    @tool
    async def loader_search(query: str) -> str:
        return query


def registry_agent_factory(config: AgentConfig) -> Agent:
    return FactoryLoadedAgent(config.name)


def import_path_agent_factory(config: AgentConfig) -> Agent:
    return FactoryLoadedAgent(config.name)


def module_agent_factory(config: AgentConfig) -> Agent:
    return FactoryLoadedAgent(config.name)


def noarg_agent_factory() -> Agent:
    return NoArgFactoryAgent("noarg")


def keyword_only_agent_factory(*, config: AgentConfig) -> Agent:
    return FactoryLoadedAgent(config.name)


def invalid_named_agent_factory(settings: AgentConfig) -> Agent:
    return FactoryLoadedAgent(settings.name)


def invalid_two_arg_agent_factory(config: AgentConfig, extra: str) -> Agent:
    return FactoryLoadedAgent(config.name)


def test_load_agent_from_config_builds_agent(tmp_path):
    _register_loader_tools()
    config_path = tmp_path / "agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: loader_bot
            model: claude-sonnet
            system_prompt: You are loaded from YAML
            tools:
              - loader_search
            memory: session
            observe: console
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, Agent)
    assert agent.name == "loader_bot"
    assert agent._config.model == "claude-sonnet"
    assert agent._config.system_prompt == "You are loaded from YAML"
    assert [tool.name for tool in agent._config.tools] == ["loader_search"]
    assert isinstance(agent._config.memories[0], SessionMemory)
    assert isinstance(agent._config.observers[0], ConsoleExporter)


def test_load_agent_config_builds_typed_config(tmp_path):
    _register_loader_tools()
    config_path = tmp_path / "agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: loader_bot
            model: claude-sonnet
            system_prompt: You are loaded from YAML
            tools:
              - loader_search
            memory: session
            observe: console
            """
        ).strip()
    )

    config = load_agent_config(str(config_path))

    assert isinstance(config, AgentConfig)
    assert config.name == "loader_bot"
    assert config.model == "claude-sonnet"
    assert config.system_prompt == "You are loaded from YAML"
    assert [tool.name for tool in config.tools] == ["loader_search"]
    assert isinstance(config.memories[0], SessionMemory)
    assert isinstance(config.observers[0], ConsoleExporter)


def test_agent_from_definition_uses_typed_config():
    config = AgentConfig(name="typed_bot", model="claude-sonnet", system_prompt="typed")

    agent = Agent.from_definition(config)

    assert isinstance(agent, Agent)
    assert agent.name == "typed_bot"
    assert agent._config is config


def test_load_agent_from_config_resolves_registry_agent_class(tmp_path):
    agent_class_registry.register("loader_custom_agent", RegistryLoadedAgent)
    config_path = tmp_path / "registry_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: registry_bot
            agent_class: loader_custom_agent
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, RegistryLoadedAgent)
    assert agent._config.agent_class == "loader_custom_agent"


def test_load_agent_from_config_resolves_import_path_agent_class(tmp_path):
    config_path = tmp_path / "import_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: import_bot
            agent_class: tests.test_config_loader.ImportPathAgent
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, ImportPathAgent)


def test_load_agent_from_config_resolves_agent_module_and_short_class(tmp_path):
    config_path = tmp_path / "module_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: module_bot
            agent_module: tests.test_config_loader
            agent_class: ModuleLoadedAgent
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, ModuleLoadedAgent)


def test_load_agent_from_config_resolves_registry_agent_factory(tmp_path):
    agent_factory_registry.register("loader_custom_factory", registry_agent_factory)
    config_path = tmp_path / "registry_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: registry_factory_bot
            agent_factory: loader_custom_factory
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, FactoryLoadedAgent)
    assert agent._config.agent_factory == "loader_custom_factory"


def test_load_agent_from_config_resolves_import_path_agent_factory(tmp_path):
    config_path = tmp_path / "import_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: import_factory_bot
            agent_factory: tests.test_config_loader.import_path_agent_factory
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, FactoryLoadedAgent)


def test_load_agent_from_config_resolves_agent_module_and_short_factory(tmp_path):
    config_path = tmp_path / "module_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: module_factory_bot
            agent_module: tests.test_config_loader
            agent_factory: module_agent_factory
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, FactoryLoadedAgent)


def test_load_agent_from_config_accepts_noarg_agent_factory(tmp_path):
    config_path = tmp_path / "noarg_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: noarg_factory_bot
            agent_factory: tests.test_config_loader.noarg_agent_factory
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, NoArgFactoryAgent)
    assert agent.name == "noarg_factory_bot"


def test_load_agent_from_config_accepts_keyword_only_config_factory(tmp_path):
    config_path = tmp_path / "kw_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: kw_factory_bot
            agent_factory: tests.test_config_loader.keyword_only_agent_factory
            model: mock
            """
        ).strip()
    )

    agent = load_agent_from_config(str(config_path), agent_cls=Agent)

    assert isinstance(agent, FactoryLoadedAgent)
    assert agent.name == "kw_factory_bot"


def test_load_agent_from_config_rejects_factory_with_wrong_parameter_name(tmp_path):
    config_path = tmp_path / "invalid_named_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: invalid_named_factory_bot
            agent_factory: tests.test_config_loader.invalid_named_agent_factory
            model: mock
            """
        ).strip()
    )

    with pytest.raises(ConfigError, match="must name its only parameter 'config'"):
        load_agent_from_config(str(config_path), agent_cls=Agent)


def test_load_agent_from_config_rejects_factory_with_multiple_parameters(tmp_path):
    config_path = tmp_path / "invalid_arity_factory_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: invalid_arity_factory_bot
            agent_factory: tests.test_config_loader.invalid_two_arg_agent_factory
            model: mock
            """
        ).strip()
    )

    with pytest.raises(ConfigError, match="either no arguments or exactly one 'config' argument"):
        load_agent_from_config(str(config_path), agent_cls=Agent)


def test_build_agent_config_rejects_unknown_tool():
    with pytest.raises(ToolNotFoundError, match="nonexistent_tool"):
        build_agent_config({"name": "loader_bot", "tools": ["nonexistent_tool"]})


def test_load_agent_from_config_rejects_unknown_tool(tmp_path):
    config_path = tmp_path / "bad_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: loader_bot
            tools:
              - nonexistent_tool
            """
        ).strip()
    )

    with pytest.raises(ToolNotFoundError, match="nonexistent_tool"):
        load_agent_from_config(str(config_path), agent_cls=Agent)


def test_load_agent_config_rejects_unknown_top_level_field(tmp_path):
    config_path = tmp_path / "bad_agent.yaml"
    config_path.write_text(
        textwrap.dedent(
            """
            name: loader_bot
            unsupported: true
            """
        ).strip()
    )

    with pytest.raises(ConfigError, match="Unknown agent config fields"):
        load_agent_config(str(config_path))



"""Stable production tests for semver, observability, reliability, and security."""
from __future__ import annotations

from datetime import UTC, datetime

import pytest

from logos_adk.agent import Agent
from logos_adk import API_VERSION, PUBLIC_API_MODULES, STABILITY_POLICY
from logos_adk.observe import InMemoryExporter, NativeObservabilityStore, Span, get_observability_store
from logos_adk.providers.mock import MockProvider
from logos_adk.scaling import SQLiteTaskQueue
from logos_adk.security import AuditLogger, DeletionRequest
from logos_adk.types import Response, Usage
from logos_adk.version import __version__


@pytest.mark.asyncio
async def test_agent_spans_include_version_and_config_metadata():
    exporter = InMemoryExporter()
    agent = Agent("bot", model=MockProvider(responses=["ok"])).observe(exporter)

    result = await agent.run("hello", session_id="s1")

    assert result.content == "ok"
    root_span = next(span for span in exporter.spans if span.name == "Agent.run")
    assert root_span.attributes["framework.version"] == __version__
    assert root_span.attributes["framework.api_version"] == API_VERSION
    assert root_span.attributes["agent.config.path"] == "inline"
    assert len(root_span.attributes["agent.config.fingerprint"]) == 16


@pytest.mark.asyncio
async def test_agent_error_span_carries_exception_type():
    class BrokenProvider(MockProvider):
        async def generate(self, messages, tools=None, output_schema=None, **kwargs):
            raise RuntimeError("boom")

    exporter = InMemoryExporter()
    agent = Agent("bot", model=BrokenProvider()).observe(exporter)

    with pytest.raises(RuntimeError, match="boom"):
        await agent.run("hello", session_id="s1")

    root_span = [span for span in exporter.spans if span.name == "Agent.run"][-1]
    assert root_span.status == "error"
    assert root_span.attributes["exception_type"] == "RuntimeError"


@pytest.mark.asyncio
async def test_sqlite_task_queue_supports_idempotency_and_dead_letter_replay(tmp_path):
    queue = SQLiteTaskQueue(path=str(tmp_path / "tasks.db"))
    calls = 0

    async def work() -> Response:
        nonlocal calls
        calls += 1
        return Response(
            content=f"result-{calls}",
            tool_calls=[],
            usage=Usage(input_tokens=1, output_tokens=1),
            raw={},
        )

    first = await queue.submit("bot", work, idempotency_key="same-key")
    second = await queue.submit("bot", work, idempotency_key="same-key")

    assert first.content == "result-1"
    assert second.content == "result-1"
    assert calls == 1

    failures = 0

    async def always_fail() -> Response:
        nonlocal failures
        failures += 1
        raise RuntimeError("fail")

    with pytest.raises(RuntimeError, match="fail"):
        await queue.submit("bot", always_fail, max_retries=1)

    stats = queue.stats()
    assert stats["failed"] == 1
    assert stats["dead_letter"] == 1
    assert failures == 2

    replay = await queue.replay(
        "dead-letter-task",
        "bot",
        work,
        max_retries=0,
    )
    assert replay.content == "result-2"


def test_observability_dashboard_and_summary_cover_new_failure_modes():
    store = NativeObservabilityStore()
    now = datetime.now(UTC)
    for exc_type in ("ProviderTimeoutError", "CircuitBreakerOpenError", "StateBackendError"):
        store.record_span(
            Span(
                name="Agent.run",
                trace_id=f"trace-{exc_type}",
                span_id=f"span-{exc_type}",
                parent_span_id=None,
                started_at=now,
                ended_at=now,
                duration_ms=10.0,
                agent_name="bot",
                session_id="s1",
                status="error",
                attributes={"exception_type": exc_type},
            )
        )

    summary = store.summary()
    assert summary["totals"]["provider_timeouts_total"] == 1.0
    assert summary["totals"]["circuit_breaker_open_total"] == 1.0
    assert summary["totals"]["state_backend_failures_total"] == 1.0

    targets = {panel["target"] for panel in store.grafana_dashboard()["panels"]}
    assert "logos_adk_provider_timeouts_total" in targets
    assert "logos_adk_circuit_breaker_open_total" in targets
    assert "logos_adk_state_backend_failures_total" in targets


def test_audit_logger_redacts_pii_and_records_deletion_requests():
    store = get_observability_store()
    store.reset()
    logger = AuditLogger()
    logger.log(
        event="pii_detection",
        outcome="warning",
        attributes={"findings": [{"pii_type": "email", "match": "secret@example.com"}]},
    )
    logger.record_deletion_request(
        DeletionRequest(
            subject_id="customer-1",
            scope="workspace:acme",
            requested_at="2026-03-18T00:00:00+00:00",
        )
    )

    logs = store.latest_logs(limit=2)
    assert logs[1]["attributes"]["findings"][0]["match"] == "[REDACTED]"
    assert logs[0]["attributes"]["event"] == "deletion_request"


def test_stable_api_surface_is_explicit():
    assert STABILITY_POLICY == "semver"
    assert "logos_adk" in PUBLIC_API_MODULES



"""Tests for artifact sandbox and toolkit."""
from __future__ import annotations

import asyncio
import json
from pathlib import Path
import zipfile


from logos_adk.agent import Agent
from logos_adk.artifacts import ArtifactSandbox
from logos_adk.context import RunContext, use_run_context


def test_artifact_sandbox_persists_text_and_json_records(tmp_path):
    sandbox = ArtifactSandbox(tmp_path / "artifacts")

    text_record = sandbox.create_text_artifact(
        workspace_id="acme",
        name="brief",
        content="launch plan",
        created_by="writer",
    )
    json_record = sandbox.create_json_artifact(
        workspace_id="acme",
        name="metrics",
        payload={"ctr": 0.42},
        created_by="analyst",
    )

    listed = sandbox.list_artifacts(workspace_id="acme")

    assert {item.id for item in listed} == {text_record.id, json_record.id}
    assert sandbox.read_text(text_record.id) == "launch plan"
    assert json.loads(sandbox.read_text(json_record.id)) == {"ctr": 0.42}


def test_create_excel_artifact_writes_valid_xlsx_archive(tmp_path):
    sandbox = ArtifactSandbox(tmp_path / "artifacts")

    record = sandbox.create_excel_artifact(
        workspace_id="acme",
        name="weekly_report",
        rows=[
            {"campaign": "spring", "ctr": 0.55},
            {"campaign": "summer", "ctr": 0.61},
        ],
        created_by="analyst",
    )

    path = sandbox.artifact_path(record.id)
    assert path is not None
    assert path.suffix == ".xlsx"
    with zipfile.ZipFile(path, "r") as archive:
        assert "xl/workbook.xml" in archive.namelist()
        sheet_xml = archive.read("xl/worksheets/sheet1.xml").decode("utf-8")
    assert "campaign" in sheet_xml
    assert "spring" in sheet_xml
    assert "0.61" in sheet_xml


def test_agent_artifact_toolkit_uses_workspace_run_context(tmp_path):
    agent = Agent("designer").artifact_sandbox(root=str(tmp_path / "artifacts"))
    tools = {tool.name: tool for tool in agent._config.tools}
    assert {
        "create_text_artifact",
        "generate_diffusion_image_artifact",
        "list_artifacts",
        "read_artifact_text",
    } <= set(tools)

    with use_run_context(RunContext(workspace_id="workspace-42", session_id="s-1")):
        created = asyncio.run(tools["create_text_artifact"](name="notes", content="hello world"))
        listed = asyncio.run(tools["list_artifacts"]())
        text = asyncio.run(tools["read_artifact_text"](artifact_id=created["id"]))

    assert created["workspace_id"] == "workspace-42"
    assert listed[0]["id"] == created["id"]
    assert text == "hello world"


def test_create_video_artifact_uses_ffmpeg_output(tmp_path, monkeypatch):
    sandbox = ArtifactSandbox(tmp_path / "artifacts")
    frame = sandbox.create_text_artifact(
        workspace_id="acme",
        name="frame",
        content="frame placeholder",
        mime_type="image/png",
        extension=".png",
    )

    monkeypatch.setattr("logos_adk.artifacts.shutil.which", lambda name: "/usr/bin/ffmpeg")

    def fake_run(command, capture_output, text, check):
        output_path = Path(command[-1])
        output_path.write_bytes(b"fake-mp4")

        class Result:
            returncode = 0
            stdout = ""
            stderr = ""

        return Result()

    monkeypatch.setattr("logos_adk.artifacts.subprocess.run", fake_run)

    record = sandbox.create_video_artifact(
        workspace_id="acme",
        name="promo_video",
        artifact_ids=[frame.id],
    )

    assert record.kind == "video"
    assert record.mime_type == "video/mp4"
    assert sandbox.artifact_path(record.id) is not None



"""Tests for RAG vector stores."""

import pytest

from logos_adk.rag import Chunk
from logos_adk.rag.vector_chroma import ChromaVectorStore
from logos_adk.rag.vector_postgresql import PostgreSQLVectorStore
from logos_adk.rag.vector_sqlite import SQLiteVectorStore
from logos_adk.rag.vector_store import InMemoryVectorStore


def _chunk(content="text", doc_id="d1", embedding=None, **meta):
    return Chunk(content=content, metadata=meta, document_id=doc_id, embedding=embedding)


class TestInMemoryVectorStore:
    @pytest.mark.asyncio
    async def test_add_and_search(self):
        store = InMemoryVectorStore()
        c = _chunk("hello", embedding=[1.0, 0.0, 0.0])
        await store.add([c])
        results = await store.search([1.0, 0.0, 0.0], limit=1)
        assert len(results) == 1
        assert results[0].content == "hello"

    @pytest.mark.asyncio
    async def test_search_empty(self):
        store = InMemoryVectorStore()
        results = await store.search([1.0, 0.0], limit=5)
        assert results == []

    @pytest.mark.asyncio
    async def test_search_returns_most_similar(self):
        store = InMemoryVectorStore()
        c1 = _chunk("close", doc_id="d1", embedding=[1.0, 0.0, 0.0])
        c2 = _chunk("far", doc_id="d2", embedding=[0.0, 1.0, 0.0])
        await store.add([c1, c2])
        results = await store.search([0.9, 0.1, 0.0], limit=1)
        assert results[0].content == "close"

    @pytest.mark.asyncio
    async def test_search_limit(self):
        store = InMemoryVectorStore()
        chunks = [_chunk(f"c{i}", doc_id=f"d{i}", embedding=[float(i), 0.0]) for i in range(5)]
        await store.add(chunks)
        results = await store.search([4.0, 0.0], limit=2)
        assert len(results) == 2

    @pytest.mark.asyncio
    async def test_search_with_filter(self):
        store = InMemoryVectorStore()
        c1 = _chunk("a", doc_id="d1", embedding=[1.0, 0.0], topic="science")
        c2 = _chunk("b", doc_id="d2", embedding=[1.0, 0.0], topic="art")
        await store.add([c1, c2])
        results = await store.search([1.0, 0.0], limit=5, filter={"topic": "science"})
        assert len(results) == 1
        assert results[0].content == "a"

    @pytest.mark.asyncio
    async def test_delete_by_document_id(self):
        store = InMemoryVectorStore()
        c1 = _chunk("a", doc_id="d1", embedding=[1.0, 0.0])
        c2 = _chunk("b", doc_id="d1", embedding=[0.0, 1.0])
        c3 = _chunk("c", doc_id="d2", embedding=[1.0, 1.0])
        await store.add([c1, c2, c3])
        await store.delete("d1")
        results = await store.search([1.0, 1.0], limit=10)
        assert len(results) == 1
        assert results[0].document_id == "d2"

    @pytest.mark.asyncio
    async def test_clear(self):
        store = InMemoryVectorStore()
        await store.add([_chunk("a", embedding=[1.0])])
        await store.clear()
        results = await store.search([1.0], limit=10)
        assert results == []

    @pytest.mark.asyncio
    async def test_skips_chunks_without_embedding(self):
        store = InMemoryVectorStore()
        c = _chunk("no vec", embedding=None)
        await store.add([c])
        results = await store.search([1.0, 0.0], limit=10)
        assert results == []


class TestSQLiteVectorStore:
    @pytest.mark.asyncio
    async def test_add_and_search(self, tmp_path):
        store = SQLiteVectorStore(str(tmp_path / "vec.db"))
        c = _chunk("hello", embedding=[1.0, 0.0, 0.0])
        await store.add([c])
        results = await store.search([1.0, 0.0, 0.0], limit=1)
        assert len(results) == 1
        assert results[0].content == "hello"

    @pytest.mark.asyncio
    async def test_persistence(self, tmp_path):
        path = str(tmp_path / "vec.db")
        store1 = SQLiteVectorStore(path)
        await store1.add([_chunk("persist", embedding=[1.0, 0.0])])

        store2 = SQLiteVectorStore(path)
        results = await store2.search([1.0, 0.0], limit=1)
        assert len(results) == 1
        assert results[0].content == "persist"

    @pytest.mark.asyncio
    async def test_delete(self, tmp_path):
        store = SQLiteVectorStore(str(tmp_path / "vec.db"))
        await store.add(
            [
                _chunk("a", doc_id="d1", embedding=[1.0, 0.0]),
                _chunk("b", doc_id="d2", embedding=[0.0, 1.0]),
            ]
        )
        await store.delete("d1")
        results = await store.search([1.0, 0.0], limit=10)
        assert len(results) == 1
        assert results[0].document_id == "d2"

    @pytest.mark.asyncio
    async def test_clear(self, tmp_path):
        store = SQLiteVectorStore(str(tmp_path / "vec.db"))
        await store.add([_chunk("a", embedding=[1.0])])
        await store.clear()
        results = await store.search([1.0], limit=10)
        assert results == []

    @pytest.mark.asyncio
    async def test_filter(self, tmp_path):
        store = SQLiteVectorStore(str(tmp_path / "vec.db"))
        c1 = _chunk("a", doc_id="d1", embedding=[1.0, 0.0], topic="science")
        c2 = _chunk("b", doc_id="d2", embedding=[1.0, 0.0], topic="art")
        await store.add([c1, c2])
        results = await store.search([1.0, 0.0], limit=5, filter={"topic": "science"})
        assert len(results) == 1
        assert results[0].metadata["topic"] == "science"


class _FakePostgresCursor:
    def __init__(self, rows):
        self._rows = rows
        self._fetchall_result = []

    def execute(self, query, params=None):
        normalized = " ".join(query.lower().split())
        if normalized.startswith("insert into"):
            chunk_id, document_id, content, metadata_json, embedding_json = params
            self._rows[chunk_id] = (
                chunk_id,
                document_id,
                content,
                metadata_json,
                embedding_json,
            )
            return
        if normalized.startswith("select id, document_id, content, metadata, embedding"):
            self._fetchall_result = list(self._rows.values())
            return
        if "where document_id = %s" in normalized:
            document_id = params[0]
            for chunk_id, row in list(self._rows.items()):
                if row[1] == document_id:
                    self._rows.pop(chunk_id, None)
            return
        if normalized.startswith("delete from"):
            self._rows.clear()
            return

    def fetchall(self):
        return self._fetchall_result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class _FakePostgresConnection:
    def __init__(self, rows):
        self._rows = rows
        self.commits = 0

    def cursor(self):
        return _FakePostgresCursor(self._rows)

    def commit(self):
        self.commits += 1

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False


class TestPostgreSQLVectorStore:
    def test_constructor_resolves_env_dsn(self, monkeypatch):
        monkeypatch.setenv("DATABASE_URL", "postgresql://env/rag")
        store = PostgreSQLVectorStore()
        assert store.dsn == "postgresql://env/rag"

    @pytest.mark.asyncio
    async def test_add_search_delete_clear_with_fake_backend(self, monkeypatch):
        rows = {}
        connection = _FakePostgresConnection(rows)
        monkeypatch.setattr(
            "logos_adk.rag.vector_postgresql.pooled_connect",
            lambda dsn: connection,
        )

        store = PostgreSQLVectorStore("postgresql://user:pass@localhost/rag")
        await store.add(
            [
                _chunk("a", doc_id="d1", embedding=[1.0, 0.0], topic="science"),
                _chunk("b", doc_id="d2", embedding=[0.0, 1.0], topic="art"),
            ]
        )

        filtered = await store.search([1.0, 0.0], limit=5, filter={"topic": "science"})
        assert len(filtered) == 1
        assert filtered[0].content == "a"

        await store.delete("d1")
        remaining = await store.search([0.0, 1.0], limit=5)
        assert len(remaining) == 1
        assert remaining[0].document_id == "d2"

        await store.clear()
        assert await store.search([1.0, 0.0], limit=5) == []


class TestChromaVectorStore:
    def test_constructor_stores_collection_name(self):
        store = ChromaVectorStore(collection_name="test_docs")
        assert store._collection_name == "test_docs"
        assert store._client is None




from typing import List, Optional
from logos_adk.tools import tool
from logos_adk.types import ToolResult

@tool
async def ask_client(question: str, options: Optional[List[str]] = None):
    """
    Остановить выполнение и задать вопрос человеку. 
    Используй это, когда тебе нужно принять решение вместе с клиентом или получить важные данные.
    Если передать список 'options', у пользователя появятся кнопки выбора.
    """
    return ToolResult(
        call_id="ask_human_pause",
        content=question,
        status="paused" # Это магический флаг, который активирует кнопки в UI
    )



"""Retrieval learning primitives for adaptive RAG tuning.

DEPRECATED: This module has been refactored into a package.
Import from logos_adk.learning.retrieval_learning instead.
"""
from __future__ import annotations

# Re-export from new package structure for backward compatibility
from .retrieval_learning import (
    QueryRewriteRecord,
    ReindexRequest,
    ReindexTrigger,
    RetrievalEvent,
    RetrievalLearningEngine,
    RetrievalProfile,
    RetrievalSettings,
    SourceDocumentSnapshot,
    SourceUsefulnessScore,
    utcnow,
)

RETRIEVAL_LEARNING_SCHEMA_KIND = "retrieval_learning"
RETRIEVAL_LEARNING_SCHEMA_VERSION = 1

__all__ = [
    "RetrievalLearningEngine",
    "RetrievalSettings",
    "QueryRewriteRecord",
    "RetrievalEvent",
    "SourceUsefulnessScore",
    "ReindexTrigger",
    "ReindexRequest",
    "SourceDocumentSnapshot",
    "RetrievalProfile",
    "RETRIEVAL_LEARNING_SCHEMA_KIND",
    "RETRIEVAL_LEARNING_SCHEMA_VERSION",
    "utcnow",
]



"""Reward scoring mechanisms for learning."""
from __future__ import annotations

from dataclasses import asdict
from json import loads
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

if TYPE_CHECKING:
    from logos_adk.learning.models import LearningExample, OutcomeRecord, RunRecord, FeedbackRecord

from .models import RewardComponent, RewardScore, LearnedPreferenceModelState
from .utils import (
    _clip_unit,
    _mean,
    _preference_target,
    _reward_feature_values,
    _sigmoid,
)


@runtime_checkable
class RewardScorer(Protocol):
    """Shared interface for heuristic and learned reward scorers."""

    def score_example(self, example: "LearningExample") -> RewardScore:
        """Return a unified reward score for a single example."""
        ...

    def score_run(
        self,
        run: "RunRecord",
        *,
        feedback: list["FeedbackRecord"],
        outcomes: list["OutcomeRecord"],
    ) -> RewardScore:
        """Return a unified reward score for a captured run."""
        ...


class HeuristicRewardScorer:
    """Aggregate human feedback, outcomes, cost, latency, and safety into one score."""

    def __init__(
        self,
        *,
        human_weight: float = 0.45,
        outcome_weight: float = 0.35,
        cost_weight: float = 0.08,
        latency_weight: float = 0.06,
        safety_weight: float = 0.06,
    ) -> None:
        self.human_weight = human_weight
        self.outcome_weight = outcome_weight
        self.cost_weight = cost_weight
        self.latency_weight = latency_weight
        self.safety_weight = safety_weight

    def score_example(self, example: "LearningExample") -> RewardScore:
        """Score a learning example using heuristic weights."""
        features = _reward_feature_values(example)
        components = [
            RewardComponent("human_feedback", features["human_feedback"], self.human_weight, features["human_feedback"] * self.human_weight),
            RewardComponent("business_outcome", features["business_outcome"], self.outcome_weight, features["business_outcome"] * self.outcome_weight),
            RewardComponent("cost_penalty", features["cost_penalty"], -self.cost_weight, -features["cost_penalty"] * self.cost_weight),
            RewardComponent("latency_penalty", features["latency_penalty"], -self.latency_weight, -features["latency_penalty"] * self.latency_weight),
            RewardComponent("safety_penalty", features["safety_penalty"], -self.safety_weight, -features["safety_penalty"] * self.safety_weight),
        ]
        score = max(0.0, min(1.0, sum(component.contribution for component in components)))
        return RewardScore(
            score=round(score, 4),
            trace_id=example.trace_id,
            run_id=example.run_id,
            workspace_id=example.workspace_id,
            agent_name=example.agent_name,
            components=components,
        )

    def score_run(
        self,
        run: "RunRecord",
        *,
        feedback: list["FeedbackRecord"],
        outcomes: list["OutcomeRecord"],
    ) -> RewardScore:
        """Score a run using heuristic weights."""
        from logos_adk.learning.models import LearningExample
        
        example = LearningExample(
            run_id=run.id,
            trace_id=run.trace_id,
            agent_name=run.agent_name,
            input_text=run.input_text,
            output_text=run.output_text,
            created_at=run.created_at,
            session_id=run.session_id,
            workspace_id=run.workspace_id,
            prompt_version=run.prompt_version,
            model=run.model,
            config_fingerprint=run.config_fingerprint,
            feedback=feedback,
            outcomes=outcomes,
            failure_types=sorted({failure for item in feedback for failure in item.failure_types}),
            corrected_output=next((item.corrected_output for item in feedback if item.corrected_output), None),
            metadata={**run.metadata, "usage": run.usage or {}},
        )
        return self.score_example(example)


class RewardModel(HeuristicRewardScorer):
    """Backward-compatible alias for the heuristic reward model."""


class LearnedPreferenceScorer:
    """Lightweight learned preference model trained from observed preference signals."""

    MODEL_VERSION = 1
    DEFAULT_FEATURE_WEIGHTS: dict[str, float] = {
        "human_feedback": 1.0,
        "business_outcome": 0.8,
        "cost_penalty": -0.25,
        "latency_penalty": -0.2,
        "safety_penalty": -1.0,
        "correction_penalty": -0.7,
    }

    def __init__(
        self,
        *,
        bias: float = 0.0,
        feature_weights: dict[str, float] | None = None,
        training_summary: dict[str, Any] | None = None,
        fallback_scorer: RewardScorer | None = None,
    ) -> None:
        self.bias = float(bias)
        self.feature_weights = dict(feature_weights or self.DEFAULT_FEATURE_WEIGHTS)
        self.training_summary = dict(training_summary or {})
        self.fallback_scorer = fallback_scorer or HeuristicRewardScorer()

    def score_example(self, example: "LearningExample") -> RewardScore:
        """Score example using learned preference model."""
        features = _reward_feature_values(example)
        contributions: list[RewardComponent] = []
        linear_score = self.bias
        for name, weight in self.feature_weights.items():
            raw_value = features.get(name, 0.0)
            centered = raw_value - 0.5
            contribution = centered * float(weight)
            linear_score += contribution
            contributions.append(
                RewardComponent(
                    name=name,
                    raw_value=raw_value,
                    weight=float(weight),
                    contribution=round(contribution, 6),
                    metadata={"centered_value": round(centered, 6), "scorer": "learned_preference"},
                )
            )
        score = _clip_unit(_sigmoid(linear_score))
        return RewardScore(
            score=round(score, 4),
            trace_id=example.trace_id,
            run_id=example.run_id,
            workspace_id=example.workspace_id,
            agent_name=example.agent_name,
            components=contributions,
        )

    def score_run(
        self,
        run: "RunRecord",
        *,
        feedback: list["FeedbackRecord"],
        outcomes: list["OutcomeRecord"],
    ) -> RewardScore:
        """Score run using learned preference model."""
        from logos_adk.learning.models import LearningExample
        
        example = LearningExample(
            run_id=run.id,
            trace_id=run.trace_id,
            agent_name=run.agent_name,
            input_text=run.input_text,
            output_text=run.output_text,
            created_at=run.created_at,
            session_id=run.session_id,
            workspace_id=run.workspace_id,
            prompt_version=run.prompt_version,
            model=run.model,
            config_fingerprint=run.config_fingerprint,
            feedback=feedback,
            outcomes=outcomes,
            failure_types=sorted({failure for item in feedback for failure in item.failure_types}),
            corrected_output=next((item.corrected_output for item in feedback if item.corrected_output), None),
            metadata={**run.metadata, "usage": run.usage or {}},
        )
        return self.score_example(example)

    @classmethod
    def train_from_examples(
        cls,
        examples: list["LearningExample"],
        *,
        fallback_scorer: RewardScorer | None = None,
    ) -> "LearnedPreferenceScorer":
        """Train preference scorer from examples."""
        if not examples:
            return cls(fallback_scorer=fallback_scorer, training_summary={"sample_size": 0, "mode": "fallback"})

        features_by_name = sorted(cls.DEFAULT_FEATURE_WEIGHTS)
        positives = {name: [] for name in features_by_name}
        negatives = {name: [] for name in features_by_name}
        labels: list[float] = []
        preference_pairs = 0

        for example in examples:
            label = _preference_target(example)
            labels.append(label)
            features = _reward_feature_values(example)
            bucket = positives if label >= 0.5 else negatives
            certainty = max(0.05, abs(label - 0.5) * 2.0)
            for name in features_by_name:
                bucket[name].append(features.get(name, 0.0) * certainty)
            if example.corrected_output or any(item.expected_output for item in example.feedback):
                preference_pairs += 1

        feature_weights: dict[str, float] = {}
        for name in features_by_name:
            pos_mean = _mean(positives[name], default=0.0)
            neg_mean = _mean(negatives[name], default=0.0)
            weight = pos_mean - neg_mean
            if abs(weight) < 0.05:
                weight = cls.DEFAULT_FEATURE_WEIGHTS.get(name, 0.0)
            feature_weights[name] = round(weight, 4)

        mean_label = _clip_unit(_mean(labels, default=0.5))
        clipped_mean = min(0.99, max(0.01, mean_label))
        from math import log
        bias = log(clipped_mean / (1.0 - clipped_mean))
        return cls(
            bias=round(bias, 4),
            feature_weights=feature_weights,
            fallback_scorer=fallback_scorer,
            training_summary={
                "sample_size": len(examples),
                "positive_examples": sum(1 for item in labels if item >= 0.5),
                "negative_examples": sum(1 for item in labels if item < 0.5),
                "mean_label": round(mean_label, 4),
                "preference_pairs": preference_pairs,
            },
        )

    @classmethod
    def train_from_store(
        cls,
        store: Any,
        *,
        workspace_id: str | None = None,
        agent_name: str | None = None,
        limit: int = 500,
        fallback_scorer: RewardScorer | None = None,
    ) -> "LearnedPreferenceScorer":
        """Train preference scorer from learning store."""
        from logos_adk.learning.dataset import DatasetBuilder
        
        builder = DatasetBuilder(store)
        examples = builder.build(
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            include_unreviewed=False,
        )
        return cls.train_from_examples(examples, fallback_scorer=fallback_scorer)

    def export_state(self) -> LearnedPreferenceModelState:
        """Export scorer state."""
        return LearnedPreferenceModelState(
            version=self.MODEL_VERSION,
            bias=round(self.bias, 6),
            feature_weights={key: round(float(value), 6) for key, value in self.feature_weights.items()},
            training_summary=dict(self.training_summary),
        )

    @classmethod
    def from_state(
        cls,
        state: LearnedPreferenceModelState | dict[str, Any],
        *,
        fallback_scorer: RewardScorer | None = None,
    ) -> "LearnedPreferenceScorer":
        """Create scorer from exported state."""
        payload = asdict(state) if isinstance(state, LearnedPreferenceModelState) else dict(state)
        return cls(
            bias=float(payload.get("bias", 0.0) or 0.0),
            feature_weights=dict(payload.get("feature_weights") or {}),
            training_summary=dict(payload.get("training_summary") or {}),
            fallback_scorer=fallback_scorer,
        )

    def save(self, path: str) -> str:
        """Save scorer state to file."""
        from json import dumps
        from pathlib import Path
        
        destination = Path(path)
        if destination.parent and not destination.parent.exists():
            destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text(
            dumps(asdict(self.export_state()), ensure_ascii=False, indent=2),
            encoding="utf-8"
        )
        return str(destination)

    @classmethod
    def load(
        cls,
        path: str,
        *,
        fallback_scorer: RewardScorer | None = None,
    ) -> "LearnedPreferenceScorer":
        """Load scorer state from file."""
        payload = loads(Path(path).read_text(encoding="utf-8"))
        return cls.from_state(payload, fallback_scorer=fallback_scorer)


def build_reward_scorer(
    mode: str,
    *,
    store: Any | None = None,
    workspace_id: str | None = None,
    agent_name: str | None = None,
    limit: int = 500,
    fallback_scorer: RewardScorer | None = None,
) -> RewardScorer:
    """Construct a reward scorer for the requested mode."""
    normalized = str(mode or "heuristic").strip().lower()
    if normalized == "heuristic":
        return RewardModel()
    if normalized == "learned":
        if store is None:
            raise ValueError("store is required for learned reward scorer mode")
        return LearnedPreferenceScorer.train_from_store(
            store,
            workspace_id=workspace_id,
            agent_name=agent_name,
            limit=limit,
            fallback_scorer=fallback_scorer,
        )
    raise ValueError(f"Unsupported reward scorer mode: {mode}")


__all__ = [
    "RewardScorer",
    "HeuristicRewardScorer",
    "RewardModel",
    "LearnedPreferenceScorer",
    "build_reward_scorer",
]



"""Retrieval ops summary Pydantic schemas."""
from __future__ import annotations

from typing import Any

from pydantic import BaseModel


class RetrievalOpsSummaryResponse(BaseModel):
    """Admin summary for retrieval operations backlog."""

    workspace_id: str | None = None
    requests: dict[str, Any]
    triggers: dict[str, Any]
    source_documents: dict[str, Any]
    policy: dict[str, Any]
    coalescing: dict[str, Any]


__all__ = ["RetrievalOpsSummaryResponse"]



"""Actor Runtime for Logos ADK."""
from __future__ import annotations

import threading
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from logos_adk.agent import Agent


@dataclass
class Supervisor:
    """Supervision policy for agent failures."""

    on_failure: str = "restart"
    max_retries: int = 3
    on_max_retries: str = "escalate"


class _DefaultRuntimeHolder:
    """Holds the process-level default Runtime."""

    def __init__(self) -> None:
        self._default: Runtime | None = None

    def get(self) -> Runtime | None:
        return self._default

    def set(self, runtime: Runtime) -> None:
        self._default = runtime

    def get_or_create(self) -> Runtime:
        if self._default is None:
            Runtime()
        assert self._default is not None
        return self._default

    def reset(self) -> None:
        """Reset default runtime. For testing only."""
        self._default = None


_default_runtime_holder = _DefaultRuntimeHolder()


class Runtime:
    """Manages agent lifecycle, message delivery, and supervision."""

    def __init__(self, supervisor: Supervisor | None = None) -> None:
        self.supervisor = supervisor or Supervisor()
        self._agents: dict[str, Agent] = {}
        self._lock = threading.Lock()

        if _default_runtime_holder.get() is None:
            _default_runtime_holder.set(self)

    def set_as_default(self) -> None:
        """Register this runtime as the process-level default."""
        _default_runtime_holder.set(self)

    def spawn(self, agent: Agent) -> None:
        """Register an agent in this runtime."""
        with self._lock:
            if agent.name in self._agents:
                raise ValueError(f"Agent '{agent.name}' already registered in this runtime")
            self._agents[agent.name] = agent
        agent._runtime = self

    def stop(self, agent: Agent) -> None:
        """Unregister an agent from this runtime."""
        with self._lock:
            self._agents.pop(agent.name, None)
        agent._runtime = None

    def get_agent(self, name: str) -> Agent | None:
        """Get a registered agent by name."""
        with self._lock:
            return self._agents.get(name)



import asyncio
from dataclasses import dataclass

import pytest

from logos_adk.a2a import A2ARequest
from logos_adk.agent import Agent
from logos_adk.event_bus import InMemoryEventBus
from logos_adk.errors import RoutingError
from logos_adk.handler import handler
from logos_adk.providers.mock import MockProvider
from logos_adk.team import Team
from logos_adk.types import Response, Usage


def _mock_provider(text: str = "ok") -> MockProvider:
    return MockProvider(responses=[
        Response(content=text, tool_calls=[], usage=Usage(1, 1), raw={}),
    ])


class TestTeamSend:
    @pytest.mark.asyncio
    async def test_send_delivers_to_mailbox(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        team = Team("t", agents=[a, b], topology="mesh")
        await team.send("a", "b", "hello")
        msg = team._mailboxes["b"].get_nowait()
        assert msg == "hello"

    @pytest.mark.asyncio
    async def test_send_blocked_by_topology(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        team = Team("t", agents=[a, b], topology={"a": ["b"], "b": []})
        with pytest.raises(RoutingError):
            await team.send("b", "a", "nope")

    @pytest.mark.asyncio
    async def test_broadcast_sends_to_all_allowed(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        c = Agent("c", model=_mock_provider())
        team = Team("t", agents=[a, b, c], topology="mesh")
        await team.broadcast("a", "hi all")
        assert team._mailboxes["b"].get_nowait() == "hi all"
        assert team._mailboxes["c"].get_nowait() == "hi all"
        assert team._mailboxes["a"].get_nowait() is None  # not to self

    @pytest.mark.asyncio
    async def test_broadcast_respects_topology(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        c = Agent("c", model=_mock_provider())
        team = Team("t", agents=[a, b, c], topology={"a": ["b"], "b": [], "c": []})
        await team.broadcast("a", "msg")
        assert team._mailboxes["b"].get_nowait() == "msg"
        assert team._mailboxes["c"].get_nowait() is None


class TestAgentSendBroadcast:
    @pytest.mark.asyncio
    async def test_agent_send_delegates_to_team(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        Team("t", agents=[a, b], topology="mesh")
        await a.send_to("b", "message")
        # Access mailbox through Team's internal structure
        assert a._team._mailboxes["b"].get_nowait() == "message"

    @pytest.mark.asyncio
    async def test_agent_send_without_team_raises(self):
        a = Agent("a", model=_mock_provider())
        with pytest.raises(RuntimeError, match="not part of a team"):
            await a.send_to("b", "msg")

    @pytest.mark.asyncio
    async def test_agent_broadcast_delegates_to_team(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        Team("t", agents=[a, b], topology="mesh")
        await a.broadcast_msg("hi")
        assert a._team._mailboxes["b"].get_nowait() == "hi"

    def test_agent_publish_event_delegates_to_team(self):
        a = Agent("a", model=_mock_provider())
        bus = InMemoryEventBus()
        Team(
            "t",
            agents=[a],
            event_bus=bus,
            topic_policy={"a": ["need_seo"]},
        )

        published = a.publish_event("need_seo", {"prompt": "optimize"})

        assert published.topic == "need_seo"
        events = bus.list_events(topic="need_seo")
        assert len(events) == 1
        assert events[0].publisher == "a"

    def test_agent_schedule_event_delegates_to_team(self):
        a = Agent("a", model=_mock_provider())
        bus = InMemoryEventBus()
        Team(
            "t",
            agents=[a],
            event_bus=bus,
            topic_policy={"a": ["need_seo"]},
        )

        scheduled = a.schedule_event("need_seo", {"prompt": "optimize"}, campaign_id="cmp-1", delay_seconds=0.1)

        assert scheduled.topic == "need_seo"
        events = bus.list_events(topic="need_seo", campaign_id="cmp-1")
        assert len(events) == 1
        assert events[0].campaign_id == "cmp-1"

    def test_agent_schedule_task_delegates_to_team(self):
        a = Agent("a", model=_mock_provider())
        bus = InMemoryEventBus()
        Team("t", agents=[a], event_bus=bus)

        scheduled = a.schedule_task(
            "check ctr tomorrow",
            delay_seconds=0.1,
            campaign_id="cmp-task",
            session_id="session-task",
        )

        assert scheduled.topic == "__scheduled_task__:a"
        events = bus.list_events(topic="__scheduled_task__:a", campaign_id="cmp-task")
        assert len(events) == 1
        assert events[0].payload["task"] == "check ctr tomorrow"
        assert events[0].payload["source_agent"] == "a"
        assert events[0].payload["target_agent"] == "a"
        assert events[0].payload["session_id"] == "session-task"

    def test_agent_publish_event_without_team_raises(self):
        a = Agent("a", model=_mock_provider())
        with pytest.raises(RuntimeError, match="not part of a team"):
            a.publish_event("need_seo", {"prompt": "optimize"})

    def test_agent_schedule_task_without_team_raises(self):
        a = Agent("a", model=_mock_provider())
        with pytest.raises(RuntimeError, match="not part of a team"):
            a.schedule_task("check later")

    @pytest.mark.asyncio
    async def test_agent_request_a2a_returns_protocol_response(self):
        requester = Agent("requester", model=_mock_provider())
        worker = Agent("worker", model=_mock_provider("completed by worker"))
        Team("t", agents=[requester, worker], topology="mesh")

        response = await requester.request_a2a("worker", "Do the task", metadata={"priority": "high"})

        assert response.status == "completed"
        assert response.from_agent == "worker"
        assert response.to_agent == "requester"
        assert response.content == "completed by worker"
        assert response.metadata["task_id"]

    @pytest.mark.asyncio
    async def test_temporal_autonomy_tool_schedules_task(self):
        analyst = Agent("analyst", model=_mock_provider()).temporal_autonomy()
        bus = InMemoryEventBus()
        Team("growth", agents=[analyst], event_bus=bus)

        tool = next(item for item in analyst._config.tools if item.name == "schedule_task")
        result = await tool(
            task="check ctr in 24h",
            delay_seconds=3600,
            campaign_id="cmp-tool",
        )

        assert result["topic"] == "__scheduled_task__:analyst"
        assert result["campaign_id"] == "cmp-tool"
        events = bus.list_events(topic="__scheduled_task__:analyst", campaign_id="cmp-tool")
        assert len(events) == 1
        assert events[0].payload["task"] == "check ctr in 24h"


@dataclass
class ResearchTask:
    query: str


@dataclass
class WriteTask:
    data: str


class TestHandlerDispatch:
    @pytest.mark.asyncio
    async def test_handler_dispatches_typed_message(self):
        """@handler methods are invoked when mailbox receives matching type."""
        received: list[ResearchTask] = []

        class Researcher(Agent):
            @handler(ResearchTask)
            async def on_research(self, task: ResearchTask) -> None:
                received.append(task)

        r = Researcher("researcher", model=_mock_provider())
        w = Agent("writer", model=_mock_provider())
        team = Team("t", agents=[r, w], topology="mesh")

        task = ResearchTask(query="AI trends")
        await team._mailboxes["researcher"].put(task)
        await team.run_rounds()

        assert len(received) == 1
        assert received[0].query == "AI trends"

    @pytest.mark.asyncio
    async def test_handler_not_matched_falls_back_to_run(self):
        """String messages without handler fall back to agent.run()."""
        provider = MockProvider(responses=[
            Response(content="fallback", tool_calls=[], usage=Usage(1, 1), raw={}),
        ])
        a = Agent("a", model=provider)
        team = Team("t", agents=[a])
        await team._mailboxes["a"].put("plain string")
        response = await team.run_rounds()
        assert response.content == "fallback"
        assert len(provider.call_history) == 1

    @pytest.mark.asyncio
    async def test_handler_chain_between_agents(self):
        """Agent A handles a message and sends to Agent B via handler."""
        write_received: list[WriteTask] = []

        class Researcher(Agent):
            @handler(ResearchTask)
            async def on_research(self, task: ResearchTask) -> None:
                await self.send_to("writer", WriteTask(data=f"found: {task.query}"))

        class Writer(Agent):
            @handler(WriteTask)
            async def on_write(self, task: WriteTask) -> None:
                write_received.append(task)

        r = Researcher("researcher", model=_mock_provider())
        w = Writer("writer", model=_mock_provider())
        team = Team("t", agents=[r, w], topology="mesh")

        await team._mailboxes["researcher"].put(ResearchTask(query="quantum"))
        await team.run_rounds()

        assert len(write_received) == 1
        assert write_received[0].data == "found: quantum"

    @pytest.mark.asyncio
    async def test_a2a_handler_can_respond_explicitly(self):
        class Researcher(Agent):
            @handler(A2ARequest)
            async def on_request(self, task: A2ARequest) -> None:
                await self.respond_a2a(task, f"researched: {task.content}", metadata={"source": "handler"})

        requester = Agent("requester", model=_mock_provider())
        researcher = Researcher("researcher", model=_mock_provider())
        Team("t", agents=[requester, researcher], topology="mesh")

        response = await requester.request_a2a("researcher", "AI adoption")

        assert response.content == "researched: AI adoption"
        assert response.metadata["source"] == "handler"


class TestTeamRoles:
    @pytest.mark.asyncio
    async def test_kickoff_can_route_by_role(self):
        researcher = Agent("researcher", model=_mock_provider("role-routed")).role(
            "Researcher",
            goal="Find facts",
        )
        writer = Agent("writer", model=_mock_provider("writer-output")).role("Writer")
        team = Team("crew", agents=[researcher, writer], topology="mesh")

        response = await team.kickoff("Investigate", role="Researcher")

        assert response.content == "role-routed"
        assert team.get_agent_by_role("Writer") is writer

    @pytest.mark.asyncio
    async def test_delegate_can_target_role(self):
        researcher = Agent("researcher", model=_mock_provider("research done")).role("Researcher")
        writer = Agent("writer", model=_mock_provider("draft ready")).role("Writer")
        team = Team("crew", agents=[researcher, writer], topology="mesh")

        response = await team.delegate("Write the brief", role="Writer")

        assert response.content == "draft ready"


class TestTeamEventBus:
    @pytest.mark.asyncio
    async def test_run_rounds_executes_due_scheduled_task_for_self(self):
        provider = _mock_provider("ctr checked")
        analyst = Agent("analyst", model=provider)
        bus = InMemoryEventBus()
        team = Team("growth", agents=[analyst], event_bus=bus)

        team.schedule_task(
            "analyst",
            "check ctr now",
            campaign_id="cmp-self-task",
            correlation_id="corr-self-task",
        )

        response = await team.run_rounds()

        assert response.content == "ctr checked"
        assert len(provider.call_history) == 1
        messages = provider.call_history[0]["messages"]
        assert messages[-1].content == "check ctr now"
        events = bus.list_events(topic="__scheduled_task__:analyst", campaign_id="cmp-self-task")
        assert len(events) == 1
        assert events[0].status == "completed"

    @pytest.mark.asyncio
    async def test_run_rounds_executes_due_scheduled_task_for_other_agent(self):
        strategist = Agent("strategist", model=_mock_provider("unused"))
        copywriter_provider = _mock_provider("copy updated")
        copywriter = Agent("copywriter", model=copywriter_provider)
        bus = InMemoryEventBus()
        team = Team(
            "growth",
            agents=[strategist, copywriter],
            topology="mesh",
            event_bus=bus,
        )

        team.schedule_task(
            "strategist",
            "rewrite headline if ctr is low",
            to_agent="copywriter",
            campaign_id="cmp-cross-task",
            correlation_id="corr-cross-task",
        )

        response = await team.run_rounds()

        assert response.content == "copy updated"
        assert len(copywriter_provider.call_history) == 1
        messages = copywriter_provider.call_history[0]["messages"]
        assert messages[-1].content == "rewrite headline if ctr is low"
        events = bus.list_events(topic="__scheduled_task__:copywriter", campaign_id="cmp-cross-task")
        assert len(events) == 1
        assert events[0].status == "completed"

    def test_schedule_task_respects_topology(self):
        strategist = Agent("strategist", model=_mock_provider())
        copywriter = Agent("copywriter", model=_mock_provider())
        team = Team(
            "growth",
            agents=[strategist, copywriter],
            topology={"strategist": [], "copywriter": []},
            event_bus=InMemoryEventBus(),
        )

        with pytest.raises(RoutingError, match="not allowed by topology"):
            team.schedule_task(
                "strategist",
                "rewrite headline",
                to_agent="copywriter",
            )

    @pytest.mark.asyncio
    async def test_agent_subscribe_topic_before_team_is_registered_on_attach(self):
        seo = Agent("seo", model=_mock_provider("optimized copy"))
        seo.subscribe_topic("need_seo", publish_to="seo_done")
        bus = InMemoryEventBus()
        team = Team(
            "t",
            agents=[seo],
            event_bus=bus,
            topic_policy={"seo": ["need_seo", "seo_done"]},
        )

        assert seo._team is team
        assert len(team._topic_subscriptions) == 2
        assert any(item.topic == "__scheduled_task__:seo" for item in team._topic_subscriptions)
        assert any(item.topic == "need_seo" for item in team._topic_subscriptions)
        seo.publish_event(
            "need_seo",
            {"prompt": "summer sale"},
            correlation_id="corr-pre",
            run_id="run-pre",
        )

        response = await team.run_rounds()

        assert response.content == "optimized copy"
        events = bus.list_events(topic="seo_done", correlation_id="corr-pre", run_id="run-pre")
        assert len(events) == 1
        assert events[0].payload["content"] == "optimized copy"

    @pytest.mark.asyncio
    async def test_agent_subscribe_topic_after_team_registers_immediately(self):
        reviewer = Agent("reviewer", model=_mock_provider("unused"))
        bus = InMemoryEventBus()
        team = Team(
            "t",
            agents=[reviewer],
            event_bus=bus,
            topic_policy={"reviewer": ["need_review", "review_done"]},
        )

        async def handle_review(event):
            return {"approved": event.payload["score"] >= 0.8}

        reviewer.subscribe_topic("need_review", handler=handle_review, publish_to="review_done")
        assert len(team._topic_subscriptions) == 2
        assert any(item.topic == "__scheduled_task__:reviewer" for item in team._topic_subscriptions)
        assert any(item.topic == "need_review" for item in team._topic_subscriptions)

        reviewer.publish_event(
            "need_review",
            {"score": 0.9},
            correlation_id="corr-post",
            run_id="run-post",
        )

        await team.run_rounds()

        events = bus.list_events(topic="review_done", correlation_id="corr-post", run_id="run-post")
        assert len(events) == 1
        assert events[0].payload == {"approved": True}

    def test_publish_event_respects_topic_policy(self):
        a = Agent("a", model=_mock_provider())
        b = Agent("b", model=_mock_provider())
        bus = InMemoryEventBus()
        team = Team(
            "t",
            agents=[a, b],
            event_bus=bus,
            topic_policy={"a": ["need_seo"], "b": []},
        )

        published = team.publish_event("a", "need_seo", {"prompt": "optimize"})
        assert published.topic == "need_seo"

        with pytest.raises(RoutingError, match="cannot publish topic"):
            team.publish_event("b", "need_seo", {"prompt": "blocked"})

    @pytest.mark.asyncio
    async def test_run_rounds_processes_team_event_subscription_and_publishes_response(self):
        seo = Agent("seo", model=_mock_provider("optimized copy"))
        bus = InMemoryEventBus()
        team = Team(
            "t",
            agents=[seo],
            event_bus=bus,
            topic_policy={"seo": ["need_seo", "seo_done"]},
        )
        team.subscribe("seo", "need_seo", publish_to="seo_done")
        team.publish_event(
            "seo",
            "need_seo",
            {"prompt": "summer sale"},
            correlation_id="corr-team",
            run_id="run-team",
        )

        response = await team.run_rounds()

        assert response.content == "optimized copy"
        request_events = bus.list_events(topic="need_seo", correlation_id="corr-team", run_id="run-team")
        assert len(request_events) == 1
        assert request_events[0].status == "completed"
        response_events = bus.list_events(topic="seo_done", correlation_id="corr-team", run_id="run-team")
        assert len(response_events) == 1
        assert response_events[0].payload["content"] == "optimized copy"

    @pytest.mark.asyncio
    async def test_schedule_event_and_campaign_state(self):
        bus = InMemoryEventBus()

        async def handle_seo(event):
            return {"done": event.payload["prompt"]}

        seo = Agent("seo", model=_mock_provider("unused"))
        team = Team(
            "t",
            agents=[seo],
            event_bus=bus,
            topic_policy={"seo": ["need_seo", "seo_done"]},
        )
        team.subscribe("seo", "need_seo", handler=handle_seo, publish_to="seo_done")

        team.schedule_event(
            "seo",
            "need_seo",
            {"prompt": "summer sale"},
            campaign_id="cmp-summer",
            correlation_id="corr-summer",
            run_id="run-summer",
            delay_seconds=0.05,
        )

        initial_state = team.campaign_state("cmp-summer")
        assert initial_state.total_events == 1
        assert initial_state.queued_events == 1

        await team.run_rounds()
        state_before_due = team.campaign_state("cmp-summer")
        assert state_before_due.completed_events == 0

        await asyncio.sleep(0.06)
        await team.run_rounds()

        state_after_due = team.campaign_state("cmp-summer")
        assert state_after_due.total_events == 2
        assert state_after_due.completed_events == 1
        response_events = bus.list_events(topic="seo_done", campaign_id="cmp-summer")
        assert len(response_events) == 1
        assert response_events[0].campaign_id == "cmp-summer"

    @pytest.mark.asyncio
    async def test_run_rounds_processes_custom_subscription_handler(self):
        reviewer = Agent("reviewer", model=_mock_provider("unused"))
        bus = InMemoryEventBus()
        team = Team(
            "t",
            agents=[reviewer],
            event_bus=bus,
            topic_policy={"reviewer": ["need_review", "review_done"]},
        )

        async def handle_review(event):
            return {"approved": event.payload["score"] >= 0.9}

        team.subscribe("reviewer", "need_review", handler=handle_review, publish_to="review_done")
        team.publish_event(
            "reviewer",
            "need_review",
            {"score": 0.95},
            correlation_id="corr-review",
            run_id="run-review",
        )

        response = await team.run_rounds()

        assert "\"approved\": true" in response.content.lower()
        response_events = bus.list_events(topic="review_done", correlation_id="corr-review", run_id="run-review")
        assert len(response_events) == 1
        assert response_events[0].payload == {"approved": True}



"""Prompt registry Pydantic schemas."""
from __future__ import annotations

from typing import Any

from pydantic import BaseModel


class PromptCreateRequest(BaseModel):
    """Create a prompt registry entry."""

    name: str
    description: str | None = None
    base_content: str | None = None
    tags: list[str] | None = None


class PromptVersionCreateRequest(BaseModel):
    """Create a prompt version."""

    content: str
    tags: list[str] | None = None
    notes: str | None = None
    created_by: str | None = None


class PromptPromoteRequest(BaseModel):
    """Promote a prompt version for a segment."""

    version: int
    environment: str = "prod"
    workspace_id: str | None = None
    task_type: str | None = None
    user_segment: str | None = None
    experiment_id: str | None = None
    enforce_guard: bool = False


class PromptRollbackRequest(BaseModel):
    """Rollback the active prompt for a segment."""

    environment: str = "prod"
    workspace_id: str | None = None
    task_type: str | None = None
    user_segment: str | None = None


class PromptExperimentCreateRequest(BaseModel):
    """Create a prompt A/B experiment."""

    baseline_version: int
    challenger_versions: list[int]
    rollout_percentage: int = 50
    metric: str = "reward"
    environment: str = "prod"
    workspace_id: str | None = None
    task_type: str | None = None
    user_segment: str | None = None


class PromptOptimizeRequest(BaseModel):
    """Run offline prompt optimization."""

    candidates: list[str]
    workspace_id: str | None = None
    feedback_limit: int = 100
    tags: list[str] | None = None
    golden_examples: list[dict[str, Any]] | None = None


class PromptRecommendationRequest(BaseModel):
    """Generate recommendation from feedback and failures."""

    workspace_id: str | None = None
    limit: int = 100


class PromptInfoResponse(BaseModel):
    """Prompt registry state."""

    prompt: dict[str, Any]
    versions: list[dict[str, Any]]
    deployments: list[dict[str, Any]]
    recommendations: list[dict[str, Any]]


class PromptExperimentResponse(BaseModel):
    """Prompt experiment response."""

    experiment: dict[str, Any]
    report: dict[str, Any] | None = None


class PromptOptimizationResponse(BaseModel):
    """Prompt optimization result."""

    created_version: dict[str, Any]
    scores: list[dict[str, Any]]


class PromptRecommendationResponse(BaseModel):
    """Prompt recommendation response."""

    recommendation: dict[str, Any]


__all__ = [
    "PromptCreateRequest",
    "PromptVersionCreateRequest",
    "PromptPromoteRequest",
    "PromptRollbackRequest",
    "PromptExperimentCreateRequest",
    "PromptOptimizeRequest",
    "PromptRecommendationRequest",
    "PromptInfoResponse",
    "PromptExperimentResponse",
    "PromptOptimizationResponse",
    "PromptRecommendationResponse",
]



"""Database row converters for retrieval learning models."""
from __future__ import annotations

import json
import sqlite3
from typing import Any

from .models import (
    ReindexRequest,
    ReindexTrigger,
    RetrievalEvent,
    RetrievalProfile,
    SourceDocumentSnapshot,
    SourceUsefulnessScore,
)


def row_to_event(row: sqlite3.Row) -> RetrievalEvent:
    """Convert database row to RetrievalEvent."""
    return RetrievalEvent(
        id=row["id"],
        trace_id=row["trace_id"],
        session_id=row["session_id"],
        workspace_id=None if row["workspace_id"] == "__global__" else row["workspace_id"],
        task_type=None if row["task_type"] == "__all__" else row["task_type"],
        user_segment=None if row["user_segment"] == "__all__" else row["user_segment"],
        prompt_name=row["prompt_name"],
        prompt_version=row["prompt_version"],
        original_query=row["original_query"],
        effective_query=row["effective_query"],
        limit=row["limit_value"],
        filter=json.loads(row["filter_json"]) if row["filter_json"] else None,
        chunks=json.loads(row["chunks_json"]),
        quality_score=float(row["quality_score"]),
        bad_context_detected=bool(row["bad_context_detected"]),
        synced_at=row["synced_at"],
        applied_feedback_ids=json.loads(row["applied_feedback_ids_json"]),
        applied_outcome_ids=json.loads(row["applied_outcome_ids_json"]),
        created_at=row["created_at"],
    )


def row_to_source_score(row: sqlite3.Row, workspace_id: str | None) -> SourceUsefulnessScore:
    """Convert database row to SourceUsefulnessScore."""
    return SourceUsefulnessScore(
        workspace_id=workspace_id,
        source_key=row["source_key"],
        connector=row["connector"],
        helpful_count=int(row["helpful_count"]),
        harmful_count=int(row["harmful_count"]),
        accepted_count=int(row["accepted_count"]),
        rejected_count=int(row["rejected_count"]),
        usefulness_score=float(row["usefulness_score"]),
        last_updated=row["last_updated"],
    )


def row_to_reindex_trigger(row: sqlite3.Row, workspace_id: str | None) -> ReindexTrigger:
    """Convert database row to ReindexTrigger."""
    return ReindexTrigger(
        id=row["id"],
        workspace_id=workspace_id,
        source_key=row["source_key"],
        reason=row["reason"],
        severity=row["severity"],
        occurrences=int(row["occurrences"]),
        status=row["status"],
        suggested_chunk_size=row["suggested_chunk_size"],
        suggested_overlap=row["suggested_overlap"],
        metadata=json.loads(row["metadata_json"]) if row["metadata_json"] else {},
        created_at=row["created_at"],
        updated_at=row["updated_at"],
    )


def row_to_reindex_request(row: sqlite3.Row, workspace_id: str | None) -> ReindexRequest:
    """Convert database row to ReindexRequest."""
    return ReindexRequest(
        id=row["id"],
        workspace_id=workspace_id,
        source_key=row["source_key"],
        reason=row["reason"],
        requested_by=row["requested_by"],
        status=row["status"],
        metadata=json.loads(row["metadata_json"]) if row["metadata_json"] else {},
        created_at=row["created_at"],
        updated_at=row["updated_at"],
    )


def row_to_source_document(row: sqlite3.Row, workspace_id: str | None) -> SourceDocumentSnapshot:
    """Convert database row to SourceDocumentSnapshot."""
    return SourceDocumentSnapshot(
        workspace_id=workspace_id,
        source_key=row["source_key"],
        document_id=row["document_id"],
        content=row["content"],
        metadata=json.loads(row["metadata_json"]) if row["metadata_json"] else {},
        source_ref=row["source_ref"],
        updated_at=row["updated_at"],
    )


def row_to_profile(row: sqlite3.Row) -> RetrievalProfile:
    """Convert database row to RetrievalProfile."""
    return RetrievalProfile(
        workspace_id=None if row["workspace_id"] == "__global__" else row["workspace_id"],
        task_type=None if row["task_type"] == "__all__" else row["task_type"],
        user_segment=None if row["user_segment"] == "__all__" else row["user_segment"],
        limit=row["limit_value"],
        filter=json.loads(row["filter_json"]) if row["filter_json"] else {},
        success_count=int(row["success_count"]),
        failure_count=int(row["failure_count"]),
        score=0.0,  # Calculated field
        last_updated=row["last_updated"],
    )


def source_score_to_dict(score: SourceUsefulnessScore) -> dict[str, Any]:
    """Convert SourceUsefulnessScore to dictionary for export."""
    return {
        "workspace_id": score.workspace_id,
        "source_key": score.source_key,
        "connector": score.connector,
        "usefulness_score": score.usefulness_score,
        "helpful_count": score.helpful_count,
        "harmful_count": score.harmful_count,
        "accepted_count": score.accepted_count,
        "rejected_count": score.rejected_count,
        "last_updated": score.last_updated,
    }



"""Mock provider for testing."""
from __future__ import annotations

from typing import Any

from logos_adk.providers import Provider
from logos_adk.types import Response, TextChunk, Usage


class MockProvider(Provider):
    """A provider that returns pre-configured responses. For testing only."""

    def __init__(
        self,
        responses: list[str | Response] | None = None,
        stream_responses: list[str | list[str] | Response] | None = None,
    ) -> None:
        self._responses: list[str | Response] = responses or ["mock response"]
        self._index = 0
        self._stream_responses: list[str | list[str] | Response] = (
            stream_responses or list(self._responses)
        )
        self._stream_index = 0
        self.call_history: list[dict[str, Any]] = []

    async def generate(self, messages, tools=None, output_schema=None, **kwargs):
        self.call_history.append({
            "messages": messages,
            "tools": tools,
            "output_schema": output_schema,
            "kwargs": kwargs,
        })
        raw_response = self._responses[self._index]
        self._index = (self._index + 1) % len(self._responses)

        if isinstance(raw_response, Response):
            return raw_response
        return Response(
            content=raw_response,
            tool_calls=[],
            usage=Usage(input_tokens=0, output_tokens=0),
            raw={},
        )

    async def stream(self, messages, tools=None, **kwargs):
        self.call_history.append({
            "messages": messages,
            "tools": tools,
            "output_schema": None,
            "kwargs": kwargs,
            "stream": True,
        })
        raw_response = self._stream_responses[self._stream_index]
        self._stream_index = (self._stream_index + 1) % len(self._stream_responses)

        if isinstance(raw_response, Response):
            if raw_response.content:
                yield TextChunk(text=raw_response.content)
            yield raw_response
            return

        if isinstance(raw_response, str):
            yield TextChunk(text=raw_response)
            return

        for chunk in raw_response:
            yield TextChunk(text=chunk)

    async def generate_batch(
        self,
        batch,
        *,
        tools=None,
        output_schema=None,
        max_concurrency=4,
        **kwargs,
    ):
        results = []
        for messages in batch:
            results.append(
                await self.generate(
                    messages,
                    tools=tools,
                    output_schema=output_schema,
                    **kwargs,
                )
            )
        return results



"""Agent state management."""
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

from logos_adk.types import Message


@dataclass
class AgentState:
    """State of a running agent. Extensible through subclassing."""

    messages: list[Message] = field(default_factory=list)
    session_id: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)
    version: int = 0
    updated_at: datetime | None = None



"""Tool middleware chain."""
from __future__ import annotations

from typing import Callable, Awaitable

from logos_adk.types import ToolCall, ToolResult

ToolExecutor = Callable[[ToolCall], Awaitable[ToolResult]]
ToolMiddleware = Callable[[ToolCall, ToolExecutor], Awaitable[ToolResult]]


class ToolMiddlewareChain:
    """Chain of middleware wrapping tool execution."""

    def __init__(self, executor: ToolExecutor) -> None:
        self._executor = executor
        self._middlewares: list[ToolMiddleware] = []

    def use(self, middleware: ToolMiddleware) -> None:
        """Add a middleware. First added = outermost (executes first)."""
        self._middlewares.append(middleware)

    async def execute(self, call: ToolCall) -> ToolResult:
        """Execute the tool call through the middleware chain."""
        chain = self._executor
        for mw in reversed(self._middlewares):
            chain = _wrap(mw, chain)
        return await chain(call)


def _wrap(mw: ToolMiddleware, next_fn: ToolExecutor) -> ToolExecutor:
    async def wrapped(call: ToolCall) -> ToolResult:
        return await mw(call, next_fn)
    return wrapped



"""Redis cache backend."""
from __future__ import annotations

import json
from typing import Any

from logos_adk.cache import CacheBackend, CacheEntry, _cosine_similarity


class RedisCacheBackend(CacheBackend):
    """Redis-backed cache with TTL support."""

    def __init__(self, url: str = "redis://localhost:6379", prefix: str = "logos:") -> None:
        self._url = url
        self._prefix = prefix
        self._client: Any = None

    async def _ensure_client(self) -> Any:
        if self._client is None:
            try:
                import redis.asyncio as aioredis
            except ImportError:
                raise ImportError(
                    "Redis cache requires the 'redis' package. "
                    "Install it with: pip install logos-adk[cache-redis]"
                )
            self._client = aioredis.from_url(self._url)
        return self._client

    def _key(self, key: str) -> str:
        return f"{self._prefix}{key}"

    async def get(self, key: str) -> CacheEntry | None:
        client = await self._ensure_client()
        raw = await client.get(self._key(key))
        if raw is None:
            return None

        data = json.loads(raw)
        data["hit_count"] = data.get("hit_count", 0) + 1
        await client.set(self._key(key), json.dumps(data), keepttl=True)

        return CacheEntry(
            response_data=data["response_data"],
            created_at=data["created_at"],
            hit_count=data["hit_count"],
            vector=data.get("vector"),
            scope=data.get("scope"),
        )

    async def put(self, key: str, entry: CacheEntry, ttl: int | None) -> None:
        client = await self._ensure_client()
        data = {
            "response_data": entry.response_data,
            "created_at": entry.created_at,
            "hit_count": entry.hit_count,
            "vector": entry.vector,
            "scope": entry.scope,
        }
        if ttl is not None:
            await client.set(self._key(key), json.dumps(data), ex=ttl)
        else:
            await client.set(self._key(key), json.dumps(data))

    async def invalidate(self, key: str) -> None:
        client = await self._ensure_client()
        await client.delete(self._key(key))

    async def clear(self) -> None:
        client = await self._ensure_client()
        cursor = 0
        while True:
            cursor, keys = await client.scan(
                cursor=cursor, match=f"{self._prefix}*", count=1000
            )
            if keys:
                await client.delete(*keys)
            if cursor == 0:
                break

    async def get_stats(self) -> dict[str, Any]:
        client = await self._ensure_client()
        cursor = 0
        total_entries = 0
        total_hits = 0
        while True:
            cursor, keys = await client.scan(
                cursor=cursor, match=f"{self._prefix}*", count=100
            )
            for key in keys:
                raw = await client.get(key)
                if raw is None:
                    continue
                total_entries += 1
                data = json.loads(raw)
                total_hits += int(data.get("hit_count", 0) or 0)
            if cursor == 0:
                break
        return {
            "backend": "Redis",
            "total_entries": total_entries,
            "total_hits": total_hits,
        }

    async def search_similar(
        self,
        vector: list[float],
        threshold: float,
        limit: int = 1,
        scope: str | None = None,
    ) -> list[CacheEntry]:
        client = await self._ensure_client()
        results: list[tuple[float, CacheEntry]] = []

        cursor = 0
        while True:
            cursor, keys = await client.scan(
                cursor=cursor, match=f"{self._prefix}*", count=100
            )
            for k in keys:
                k_str = k.decode() if isinstance(k, bytes) else k
                if k_str.endswith(":vec"):
                    continue
                raw = await client.get(k)
                if raw is None:
                    continue
                data = json.loads(raw)
                stored_vec = data.get("vector")
                if stored_vec is None:
                    continue
                if scope is not None and data.get("scope") != scope:
                    continue
                sim = _cosine_similarity(vector, stored_vec)
                if sim >= threshold:
                    results.append((sim, CacheEntry(
                        response_data=data["response_data"],
                        created_at=data["created_at"],
                        hit_count=data.get("hit_count", 0),
                        vector=stored_vec,
                        scope=data.get("scope"),
                    )))
            if cursor == 0:
                break

        results.sort(key=lambda x: x[0], reverse=True)
        return [entry for _, entry in results[:limit]]
